"""
Base classes for all modules of the simulation loop.

Every block of the pymoskito simulation loop is implemented
by subclassing `SimulationModule`.
"""
import logging
from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from collections.abc import Callable
from copy import copy

from PyQt5.QtCore import QObject

pyqtWrapperType = type(QObject)

__all__ = ["SimulationModule", "SimulationException",
           "Trajectory", "Feedforward", "Controller", "Limiter",
           "ModelMixer", "Model", "ModelException",
           "Solver", "Disturbance", "Sensor", "ObserverMixer", "Observer"]


class SimulationModuleMeta(ABCMeta, pyqtWrapperType):
    pass


class SimulationException(Exception):
    pass


class SimulationModule(QObject, metaclass=SimulationModuleMeta):
    """
    Base unit of the simulation framework.

    This class provides necessary functions like output calculation and holds
    all settings that can be accessed by the user.
    The :py:attr:`public_settings` are read by the
    :py:class:`.SimulationInterface` and the rendered by the GUI. All entries
    stated in this dictionary will be available as editable settings for the
    module.
    On initialization, a possibly modified (in terms of its values) version of
    this dict will be passed back to this class and is thenceforward available
    via the :py:attr:`settings` property.

    The most important method is :py:func:`calc_output` which is called by the
    :py:class:`Simulator` to retrieve this modules output.

    Args:
        settings(OrderedDict): Settings for this simulation module.
            These entries will be shown in the properties view and can be
            changed by the user. The important entries for this base class are:

            `tick_divider`:
                This property controls the frequency at which this Module is evaluated
                in the main simulation loop. If e.g. a value of `2` is provided, the
                module will be evaluated at every 2nd step that the solver makes.
                This feature comes in handy if discrete setups with different sample rates
                are simulated.
            `output info`:
                Dict holding information about the model output that are used
                in the GUI. For every item in the output array, a key for its
                index can be provided, under whose value another dict with the
                keys keys `Name` and `Unit` is expected.
                If available, these information are used to display reasonable
                names in the result view and to display the corresponding units
                for the result plots.

    Note:
        The 'tick_divider' setting is ignored for the Solver and Model modules, whose `tick_divider` is
        fixed at 1.

    Warn:
        Do NOT use '.' in the `output_info` name field.

    TODO:
        Get rid of the point restriction
    """

    def __init__(self, settings):
        QObject.__init__(self, None)
        self._logger = logging.getLogger(self.__class__.__name__)

        assert isinstance(settings, dict)
        self._settings = copy(settings)

        self._settings["tick divider"] = settings.get("tick divider", 1)
        self._settings["step width"] =  settings.get("step width", None)
        self._settings.pop("modules", None)

    @property
    @abstractmethod
    def public_settings(self):
        """
        Dict of public settings.

        This contains all settings that will be shown and can be edited
        in the GUI.
        """
        pass

    @property
    def settings(self):
        return self._settings

    @property
    def tick_divider(self):
        """
        The module tick divider.
        """
        return self._settings["tick divider"]

    @property
    def step_size(self):
        """
        The step size that this module is called with.

        This parameter will be computed when the simulation is initialised
        and is based on the solver step size and the selected tick
        divider of the module.
        """
        return self._settings["step size"]

    @abstractmethod
    def calc_output(self, input_vector):
        """
        Main evaluation routine.

        Every time the solver has made the amount of steps specified in `tick_divider`,
        this method will be called to compute the next output of the module.
        This is typically where the main logic of a module is implemented.
        """
        pass


class ModelException(SimulationException):
    """
    Exception to be raised if the current system state violates modelling
    assumptions.
    """
    pass


class Model(SimulationModule):
    """
    Base class for all user defined system models in state-space form.

    Args:
        settings (OrderedDict): Dictionary holding the config options for this module.
            It *must* contain the following keys:

            :input_count:
                The length of the input vector for this model.

            :state_count:
                The length of the state vector for this model.

            :initial state:
                The initial state vector for this model.
    """

    def __init__(self, settings):
        SimulationModule.__init__(self, settings)
        assert ("state_count" in settings)
        assert ("input_count" in settings)
        assert ("initial state" in settings)
        assert len(settings["initial state"]) == settings["state_count"]
        assert self.tick_divider == 1
        self._events = []

    def register_event(self, event):
        """
        Register switching events with the model

        These events do help the solver if the system equations
        are discontinuous.
        For an example event function, see `event_example`.
        """
        if not hasattr(self, "_events"):
            raise ModelException("Custom events for the Model can only be registered AFTER "
                                 "the base class init() has been called.")
        if not isinstance(event, Callable):
            raise TypeError("A custom event must be callable, see Model.empty_event for an example.")
        self._events.append(event)

    @property
    def events(self):
        return self._events

    @property
    def initial_state(self):
        """ Return the initial state of the model. """
        return self._settings["initial state"]

    @abstractmethod
    def state_function(self, t, x, args):
        """
        Calculate the state derivatives of a system with state x at time t.

        Args:
            x(Array-like): System state.
            t(float): System time.
            args: Extra arguments.

        Returns:
            Temporal derivative of the system state at time t.
        """
        pass

    def event_example(self, t, x):
        """
        This must be a continuous function of time and state.

        Sign changes of this function are considered to be an event
        where a reinitialisation of the integrator must be performed
        because of discontinuities in the system dynamics.

        Args:
            t: Current time.
            x(array-like): Current system state.
            
        Returns: 
            Float
                
        """
        return x[0]

    def check_consistency(self, x):
        """
        Check whether the assumptions, made in the modelling process are 
        violated.
        
        Args:
            x: Current system state
            
        Raises: 
            :py:class:`ModelException` : If a violation is detected. This will
                stop the simulation process.
        """
        pass


class SolverException(SimulationException):
    """
    Exception to be raised if the solver encounters any problems
    while integrating the system.
    """
    pass


class Solver(SimulationModule):
    """
    Base Class for solver implementations.

    After initialization, for every step in the simulation
    `set_input` will be called, followed by `integrate`.

    Args:
        settings (OrderedDict): Dictionary holding the config options for this module.
            It must contain the following keys:

            :start time: Time stamp to start the simulation.

            :end time: Time stamp to end the simulation.

            :measure rate: How many results to store per second, e.g. a 'measure rate' of 500 will
            result of 500 values per second, thus one value every 2 milliseconds.

    """

    def __init__(self, settings):
        assert isinstance(settings["modules"]["Model"], Model)
        self._model = settings["modules"]["Model"]
        self.next_output = None
        SimulationModule.__init__(self, settings)
        assert self.tick_divider == 1

    def calc_output(self, input_vector):
        self.set_input(input_vector["Model_Input"])
        output = self.next_output

        self.next_output = self.integrate(t=None)
        try:
            self._model.check_consistency(self.next_output)
        except ModelException as e:
            self._logger.exception(e)
            raise SolverException("Time step integration failed!")
        return output

    @abstractmethod
    def set_input(self, *args):
        """
        Updates the inputs for the next integration step.
        """
        pass

    @abstractmethod
    def integrate(self, t):
        """
        Perform numerical integration.

        Args:
            t(float): New time up to which the model is to be integrated.
        """
        pass

    @property
    @abstractmethod
    def t(self):
        """
        The current simulation time.
        """
        pass

    @property
    @abstractmethod
    def successful(self):
        pass


class ControllerException(SimulationException):
    """
    Exception to be raised if the controller encounters any problems.
    """
    pass


class Controller(SimulationModule):
    """
    Base class for controllers.

    After subclassing, the method `_control` has to be implemented.

    Args:
        settings (OrderedDict): Dictionary holding the config options for this module.
            It must contain the following keys:

            :input_order:
                The order of required derivatives from the trajectory generator.

            :input_type:
                Source for the feedback calculation and one of the following:
                `Model_State` , `Model_Output` , `Observer` or `Sensor` .
    """
    # selectable input sources for controller
    input_sources = ["Model_State", "Model_Output", "Observer", "Sensor"]

    def __init__(self, settings):
        assert ("input_order" in settings)
        assert ("input_type" in settings)
        SimulationModule.__init__(self, settings)
        if self._settings["input_type"] == "system_state":
            self._logger.warning(
                "Input source `system_state` is being deprecated in favour of `Model_State`."
                " Please adapt your implementation.")
            self._settings["input_type"] = "Model_State"
        if self._settings["input_type"] == "system_output":
            self._logger.warning(
                "Input source `system_output` is being deprecated in favour of `Model_Output`."
                " Please adapt your implementation.")
            self._settings["input_type"] = "Model_Output"
        assert (self._settings["input_type"] in self.input_sources)

    @property
    def input_order(self):
        return self._settings["input_order"]

    def calc_output(self, input_vector):
        input_values = next((input_vector[src] for src in self.input_sources
                             if src == self._settings["input_type"]), None)
        if input_values is None:
            raise ControllerException("Selected Input not available")
        trajectory_values = input_vector.get("Trajectory", None)
        feedforward_values = input_vector.get("Feedforward", None)

        return self._control(input_vector["time"], trajectory_values,
                             feedforward_values, input_values)

    @abstractmethod
    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        """
        Placeholder for control law calculations.

        For more sophisticated implementations overload :py:func:`calc_output` .

        Args:
            time (float): Current time.
            trajectory_values (array-like): Desired values from the trajectory
                generator.
            feedforward_values (array-like): Output of feedforward block.
            input_values (array-like): The input values selected by
                ``input_type`` .
            **kwargs: Placeholder for custom parameters.

        Returns:
            Array: Control output.
        """
        pass


class Observer(SimulationModule):
    """
    Base class for observers.

    After subclassing, the method `_observe` has to be implemented.
    """

    def __init__(self, settings):
        SimulationModule.__init__(self, settings)

    def calc_output(self, input_vector):
        model_input = input_vector.get("Model_Input", None)
        if "ObserverMixer" in input_vector:
            model_output = input_vector["ObserverMixer"]
        elif "Model_Output" in input_vector:
            model_output = input_vector["Model_Output"]
        else:
            raise SimulationException("No Observer input specified")

        return self._observe(input_vector["time"], model_input, model_output)

    @abstractmethod
    def _observe(self, time, system_input, system_output):
        """
        Placeholder for observer law.

        Note:
            For integration of the observer dynamics, the attribute `step_width`
            should be used the use the correct step size.

        Args:
            time: Current time.
            system_input: Current system input.
            system_output: Current system output.

        Returns:
            Estimated system state
        """
        pass


class Feedforward(SimulationModule):
    """
    Base class for all feedforward implementations.

    After subclassing, the method `_feedforward` has to be implemented.
    """

    def __init__(self, settings):
        self._model = settings["modules"]["Model"]
        SimulationModule.__init__(self, settings)
        assert ("input_order" in settings)

    @property
    def input_order(self):
        """
        This field can be used to specify
        the derivative order of the reference trajectories that have to
        be provided by the trajectory generator.
        """
        return self._settings["input_order"]

    def calc_output(self, input_dict):
        return self._feedforward(input_dict["time"], input_dict["Trajectory"])

    @abstractmethod
    def _feedforward(self, time, trajectory_values):
        """
        Placeholder for feedforward calculations.

        Args:
            time (float): Current time.
            trajectory_values(array-like): Desired values from the trajectory
                generator.

        Returns:
            Array: Feedforward output. This signal can be added to the
            controllers output via the :py:class:`.ModelMixer` and is also
            directly passed to the controller.
        """
        pass


class TrajectoryException(SimulationException):
    """
    Exception to be raised for errors during trajectory generation.
    """
    pass


class Trajectory(SimulationModule):
    """
    Base class for all trajectory generators.
    """

    def __init__(self, settings):
        if "modules" in settings:
            control_order = 0
            feedforward_order = 0
            if "Controller" in settings["modules"].keys():
                control_order = settings["modules"]["Controller"].input_order
            if "Feedforward" in settings["modules"].keys():
                feedforward_order = settings["modules"]["Feedforward"
                ].input_order
            settings.update(differential_order=max([control_order,
                                                    feedforward_order]))
        else:
            assert "differential_order" in settings
        SimulationModule.__init__(self, settings)

    def calc_output(self, input_vector):
        desired = self._desired_values(input_vector["time"])
        return desired

    @abstractmethod
    def _desired_values(self, t):
        """
        Placeholder for calculations of desired values.

        Args:
            t (float): Time.

        Returns:
            Array: Trajectory output. This should always be a two-dimensional
            array holding the components in to 0th and their derivatives in
            the 1th axis.
        """
        pass


class MixerException(Exception):
    pass


class SignalMixer(SimulationModule):
    """
    Base class for all signal mixing modules

    Requires the key 'input signals' to be present in the settings dictionary.
    For all possible keys see :py:class:`pymoskito.simulation_core.simulator._dynamic_module_list`.
    """

    def __init__(self, settings):
        assert "input signals" in settings
        SimulationModule.__init__(self, settings)

    def calc_output(self, input_vector):
        signals = [input_vector[s]
                   for s in self._settings["input signals"]
                   if s != "None" and s is not None]
        return self._mix(signals)


class ModelMixer(SignalMixer):
    pass


class ObserverMixer(SignalMixer):
    pass


class Limiter(SimulationModule):
    """
    Base class for all limiter variants
    """

    def __init__(self, settings):
        assert "input_signal" in settings
        SimulationModule.__init__(self, settings)

    def calc_output(self, input_dict):
        return self._limit(input_dict[self._settings["input_signal"]])

    def _limit(self, values):
        """
        Placeholder for actual limit calculations.

        Args:
            values(array-like): Values to limit.

        Returns:
            Array: Limited output.
        """
        return values


class Sensor(SimulationModule):
    """
    Base class for all sensor variants
    """

    def __init__(self, settings):
        assert "input signal" in settings
        SimulationModule.__init__(self, settings)

    def calc_output(self, input_dict):
        return self._measure(input_dict[self._settings["input signal"]])

    def _measure(self, value):
        """
        Placeholder for measurement calculations.

        One may reorder or remove state elements or introduce measurement delays
        here.

        Args:
            value (array-like float): Values from the source selected by the
                ``input_signal`` property.
        Returns:
            array-like float: 'Measured' values.
        """
        return value


class Disturbance(SimulationModule):
    """
    Base class for all disturbance variants
    """

    def __init__(self, settings):
        assert "input signal" in settings
        SimulationModule.__init__(self, settings)

    def calc_output(self, input_dict):
        return self._disturb(
            input_dict["time"],
            input_dict.get(self._settings["input signal"], None))

    @abstractmethod
    def _disturb(self, t, signal):
        """
        Placeholder for disturbance calculations.

        If the noise is to be dependent on the measured signal use its `value`
        to create the noise.

        Args:
            t (float): Current simulation time.
            signal (array-like float): Values from the source selected by the
                ``input_signal`` property.
        Returns:
            array-like float: Noise that will be mixed with a signal later on.
        """
        pass



from collections import OrderedDict

import numpy as np

import pymoskito as pm
from . import settings as st


class TwoTankSystem(pm.Model):
    public_settings = OrderedDict([('initial state', st.initial_states),
                                   ("hT", st.hT),
                                   ("hV", st.hV),
                                   ("AS1", st.AS1),
                                   ("AS2", st.AS2),
                                   ("AT1", st.AT1),
                                   ("AT2", st.AT2),
                                   ("g", st.g),
                                   ("K", st.K)])

    def __init__(self, settings):  # Constructor
        # add specific "private" settings
        settings.update(state_count=2)
        settings.update(input_count=1)
        settings.update({"output_info": {
            0: {"Name": "Level Tank1", "Unit": "m"},
            1: {"Name": "Level Tank2", "Unit": "m"},
        }})
        pm.Model.__init__(self, settings)

        # register events
        self.register_event(self.event_level_1_low)
        self.register_event(self.event_level_2_low)
        self.register_event(self.event_level_1_high)
        self.register_event(self.event_level_2_high)

    def state_function(self, t, x, args):
        """
        Calculations of system state changes
        :param x: state
        :param t: time
        :type args: system input u
        """
        x1 = x[0]
        x2 = x[1]

        uA = args[0].squeeze()
        g = self.settings['g']
        K = self.settings['K']
        AS1 = self.settings['AS1']
        AS2 = self.settings['AS2']
        AT1 = self.settings['AT1']
        AT2 = self.settings['AT2']
        hT = self.settings["hT"]
        hV = self.settings["hV"]
        a1 = AS1 * np.sqrt(2 * g / (AT1 ** 2 - AS1 ** 2))
        a2 = AS2 * np.sqrt(2 * g / (AT2 ** 2 - AS2 ** 2))

        dx = np.zeros(2)
        if x1 < hT:
            # tank 1 not full, count inflow
            dx[0] += K / AT1 * uA
        if x1 >= 0 and x2 < hT:
            # tank 1 not empty and tank 2 not full, count out- and inflow
            dx[0] -= a1 * np.sqrt(x1 + hV)
            dx[1] += a1 * np.sqrt(x1 + hV)
        if x2 >= 0:
            # tank not empty, count outflow
            dx[1] -= a2 * np.sqrt(x2 + hV)

        return dx

    def event_level_1_low(self, t, x):
        return x[0]

    def event_level_2_low(self, t, x):
        return x[1]

    def event_level_1_high(self, t, x):
        return x[0] - self.settings["hT"]

    def event_level_2_high(self, t, x):
        return x[1] - self.settings["hT"]

    def calc_output(self, input_vector):
        return input_vector


pm.register_simulation_module(pm.Model, TwoTankSystem)



# -*- coding: utf-8 -*-

#mpl.rcParams['text.usetex']=True
#mpl.rcParams['text.latex.unicode']=True
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

from processing_gui import MetaProcessingModule
from tools import sortTree


class eval_L1NormAbs_delay_linePlot(MetaProcessingModule):
    '''
    create diagrams for evaluation of L1NormAbs
    '''
    def __init__(self):
        MetaProcessingModule.__init__(self)
        return

    def run(self, postResults):
        #create tree with relevant datas
        source = sortTree(postResults, ['modules', 'controller', 'type'])
        
        #create plot
        fig = Figure()
        fig.subplots_adjust(wspace=0.5, hspace=0.25)
        
        #plot for L1NormAbs
        axes = fig.add_subplot(111)
                        
        self.plotVariousController(source, axes,\
                xPath=['modules', 'sensor', 'delay'],\
                yPath=['metrics', 'L1NormAbs'],\
                typ='line')
        self.plotSettings(axes,\
                titel=r'Fehlerintegral w(t) und y(t) \"uber Delay',\
                grid=True,\
                xlabel=r'$Delay \, \lbrack ms \rbrack$',\
                ylabel=r'$E \, \lbrack ms \rbrack$',\
                )
        
        #extract controllerNames
        controllerNames = [x[:-len('Controller')] for x in source.keys()]
        
        canvas = FigureCanvas(fig)
        #write output files
        fileName = self.name[len('eval_'):]\
                    + '_Controller_(' + ''.join(controllerNames) + ')'
        self.writeOutputFiles(fileName, fig)
        
        return [{'figure': canvas, 'name': self.name}]



from collections import OrderedDict
import os
import pickle
import warnings

from scipy.integrate import ode
from scipy.interpolate import interp1d
from scipy.optimize import bisect
import sympy as sp
import numpy as np

from .simulation_modules import (
    Model, Solver, SolverException,
    Trajectory, TrajectoryException, Controller,
    Feedforward, SignalMixer, ModelMixer, ObserverMixer, Limiter, Sensor,
    Disturbance
)
from .controltools import calc_prefilter, place_siso

__all__ = ["LinearStateSpaceModel", "ODEInt", "ModelInputLimiter",
           "Setpoint", "HarmonicTrajectory", "SmoothTransition", "InterpolatorTrajectory",
           "Feedthrough",
           "PIDController", "LinearStateSpaceController",
           "DeadTimeSensor", "GaussianNoise",
           "AdditiveMixer"]

"""
Ready to go implementations of simulation modules.
"""


class LinearStateSpaceModel(Model):
    """
    The state space model of a linear system.

    The parameters of this model can be provided in form of a file whose path is
    given by the setting ``config file`` .
    This path should point to a pickled dict holding the following keys:

        `system`:
            An Instance of :py:class:`scipy.signal.StateSpace` (from scipy)
            representing the system,
        `op_inputs`:
            An array-like object holding the operational point's inputs,
        `op_outputs`:
            An array-like object holding the operational point's outputs.

    """
    public_settings = OrderedDict([
        ("config file", None),
        ("initial state", None),
        ("initial output", None),
    ])

    def __init__(self, settings):
        file = settings["config file"]
        assert os.path.isfile(file)

        with open(file, "rb") as f:
            data = pickle.load(f)

        if "system" not in data:
            raise ValueError("Config file lacks mandatory settings.")

        self.ss = data["system"]

        # no feedthrough possible
        np.testing.assert_array_equal(self.ss.D, np.zeros_like(self.ss.D))

        settings["state_count"] = self.ss.B.shape[0]
        settings["input_count"] = self.ss.B.shape[1]

        self.input_offset = data.get("op_inputs",
                                     np.zeros((self.ss.B.shape[1], )))
        if len(self.input_offset) != self.ss.B.shape[1]:
            raise ValueError("Provided input offset does not match input "
                             "dimensions.")

        self.output_offset = data.get("op_outputs",
                                      np.zeros((self.ss.C.shape[0], )))
        if len(self.output_offset) != self.ss.C.shape[0]:
            raise ValueError("Length of provided output offset does not match "
                             "output dimensions ({} != {}).".format(
                len(self.output_offset),
                self.ss.C.shape[0]
            ))

        if settings["initial state"] is None:
            if settings["initial output"] is None:
                raise ValueError("Neither 'initial state' nor 'initial output'"
                                 "given.")

            settings["initial state"] = \
                np.linalg.pinv(self.ss.C) @ (settings["initial output"]
                                             - self.output_offset)

        super().__init__(settings)

    def state_function(self, t, x, args):
        return np.squeeze(self.ss.A @ x + self.ss.B @ (
            args[0] - self.input_offset))

    def calc_output(self, input_vector):
        return (self.ss.C @ input_vector
                + self.output_offset)


class ODEInt(Solver):
    """
    Wrapper for ode_int from the Scipy project

    Args:
        settings (OrderedDict): Dictionary holding the config options for this module.
            See :py:class:`Solver` and scipy's odeint for details.

    """
    public_settings = OrderedDict([
        ("measure rate", 500),
        ("start time", 0),
        ("end time", 5),
        ("Mode", "vode"),
        ("Method", "adams"),
        ("step size", 1e-3),
        ("rTol", 1e-6),
        ("aTol", 1e-9),
    ])

    def __init__(self, settings):
        Solver.__init__(self, settings)

        # setup bisection fields
        self._cur_state = np.atleast_1d(self._model.initial_state)
        self._cur_events = self._eval_events(self._settings["start time"], self._cur_state)

        # setup solver
        if hasattr(self._model, "jacobian"):
            self._solver = ode(self._model.state_function,
                               jac=self._model.jacobian)
        else:
            self._solver = ode(self._model.state_function)

        self._solver.set_integrator(self._settings["Mode"],
                                    method=self._settings["Method"],
                                    rtol=self._settings["rTol"],
                                    atol=self._settings["aTol"],
                                    max_step=self._settings["step size"]
                                    )
        self._solver.set_initial_value(self._cur_state, t=self._settings["start time"])

    def _eval_events(self, t, x):
        """ Evaluation helper for the event function """
        # call all events
        r = np.array([f(t, x) for f in self._model.events],
                     dtype=float)
        # an actual zero is counted as positive
        r[r == 0] = np.finfo(float).eps
        return r

    @property
    def t(self):
        return self._solver.t

    @property
    def successful(self):
        return self._solver.successful()

    def set_input(self, *args):
        """
        propagate input changes to ode_int
        """
        self._solver.set_f_params(args)
        if hasattr(self._model, "jacobian"):
            # TODO Test
            self._solver.set_jac_params(args)

    def integrate(self, *, t=None):
        """
        Solve the system equations for one step

        Args:
            t (float): Target time to stop at, if None use step size from settings
        Returns:
            system state at target time
        """
        cur_time = self.t
        cur_state = self._cur_state

        if t is None:
            new_time = cur_time + self._settings["step size"]
        else:
            new_time = t
        new_state = self._solver.integrate(new_time)
        if not self._solver.successful():
            raise SolverException("Integration failed.\n"
                                  "This can happen due to several reasons, here are some pointers:\n"
                                  "- First of all: Check the log for solver warnings\n"
                                  "- Model: Check for errors and inconsistencies in the model equations\n"
                                  "- For switching model equations: Try to provide event functions to help the solver\n"
                                  "- Controller/Feedforward: Check for unreasonably large inputs\n"
                                  "- Solver: Loosen the precision requirements by increasing aTol and/or rTol"
                                  )

        # evaluate custom event functions for new state
        new_events = self._eval_events(new_time, new_state)
        # check if any events (sign changes) occurred
        e_idxs = np.where(self._cur_events * new_events < 0)[0]

        # run bisection to get exact time for each event
        e_times = []
        for e_idx in e_idxs:
            self._logger.debug(f"Event with index {e_idx} detected between "
                               f"{cur_time} and {self.t} seconds")
            try:
                e_t = bisect(self._b_func, a=cur_time, b=self.t,
                                args=(cur_time, cur_state, e_idx))
            except SolverException as err:
                e_t = self.t
            except ValueError as err:
                    if err.args[0] == 'f(a) and f(b) must have different signs':
                        self._logger.debug("Event could not be reproduced, ignoring")
                        continue
                    else:
                        raise SolverException(f"Bisection failed with error: '{err}'")
            self._logger.debug(f"Bisection yielded event time {e_t}")
            e_times.append(e_t)

        if len(e_times) > 0:
            # take the smallest time stamp and continue
            e_idx = np.argmin(e_times)
            e_time = e_times[e_idx]

            self._logger.debug(f"Restarting at original t={cur_time} from state {cur_state} "
                               f"and solving up to event at t={e_time}")
            # reset the solver to the original starting time
            self._solver.set_initial_value(cur_state, cur_time)
            # integrate up to event time
            temp_state = self._solver.integrate(e_time)
            temp_events = self._eval_events(e_time, temp_state)
            # make sure we are on the right side of the change
            temp_events[e_idx] = np.copysign(temp_events[e_idx], new_events[e_idx])

            self._logger.debug(f"Restarting at event t={e_time} from state {temp_state} "
                               f"and solving up to desired t={new_time}")
            # reset the solver to the event time
            self._solver.set_initial_value(temp_state, e_time)
            # integrate up to the desired time
            self._cur_state = temp_state
            self._cur_events = temp_events
            new_state = self.integrate(t=new_time)
            self._logger.debug(f"Event handled")

        # update internal fields
        self._cur_state = new_state
        self._cur_events = new_events

        return new_state

    def _b_func(self, t_end, t_start, x_start, e_idx):
        """
        Helper function for bisection of the exact event time
        """
        self._solver.set_initial_value(x_start, t_start)
        x_end = self._solver.integrate(t_end)
        if not self._solver.successful():
            raise SolverException("Solver failed during bisection")
        e_end = self._eval_events(t_end, x_end)
        return e_end[e_idx]


class SmoothTransition(Trajectory):
    """
    provides (differential) smooth transition between two scalar states
    """
    public_settings = {"states": [[0, 1]],
                       "start time": 0,
                       "delta t": 5,
                       }

    def __init__(self, settings):
        settings["states"] = np.asarray(settings["states"])
        Trajectory.__init__(self, settings)

        # setup symbolic expressions
        tau, k = sp.symbols("tau, k")

        gamma = self._settings["differential_order"] + 1
        alpha = sp.factorial(2 * gamma + 1)

        f = sp.binomial(gamma, k) * (-1) ** k * tau ** (gamma + k + 1) / (gamma + k + 1)
        phi = alpha / sp.factorial(gamma) ** 2 * sp.summation(f, (k, 0, gamma))

        # differentiate phi(tau), index in list corresponds to order
        dphi_sym = [phi]  # init with phi(tau)
        for order in range(self._settings["differential_order"]):
            dphi_sym.append(dphi_sym[-1].diff(tau))

        # lambdify
        self.dphi_num = []
        for der in dphi_sym:
            self.dphi_num.append(sp.lambdify(tau, der, "numpy"))

        # issue deprecation warning
        if self._settings["states"].ndim == 1:
            msg = "SmoothTransition will require 2d data for 'states' " \
                  "in the next version, please update your configuration."
            warnings.warn(msg, DeprecationWarning)
            self._logger.warn(msg)
            self._settings["states"] = self._settings["states"][None, :]

    def _desired_values(self, t):
        """
        Calculates desired trajectory
        """
        yd = self._settings['states']
        t0 = self._settings['start time']
        dt = self._settings['delta t']

        y = np.zeros((yd.shape[0], len(self.dphi_num)))
        if t < t0:
            y[:, 0] = yd[:, 0]
        elif t > t0 + dt:
            y[:, 0] = yd[:, 1]
        else:
            for order, dphi in enumerate(self.dphi_num):
                if order == 0:
                    ya = yd[:, 0]
                else:
                    ya = np.zeros_like(yd[:, 0])

                scale = dphi((t - t0) / dt) * 1 / dt ** order
                y[:, order] = ya + (yd[:, 1] - yd[:, 0]) * scale

        return y


class HarmonicTrajectory(Trajectory):
    """
    This generator provides a scalar harmonic sinus signal
    with derivatives up to order n
    """

    public_settings = OrderedDict([("Amplitude", 0.25),
                                   ("Frequency", 0.1),
                                   ("Offset", 0.75),
                                   ("Phase in degree", 0)])

    def __init__(self, settings):
        Trajectory.__init__(self, settings)

        # calculate symbolic derivatives up to order n
        t, a, f, off, p = sp.symbols("t, a, f, off, p")
        self.yd_sym = []
        harmonic = a * sp.sin(2 * sp.pi * f * t + p) + off

        for idx in range(settings["differential_order"] + 1):
            self.yd_sym.append(harmonic.diff(t, idx))

        # lambdify
        for idx, val in enumerate(self.yd_sym):
            self.yd_sym[idx] = sp.lambdify((t, a, f, off, p), val, "numpy")

    def _desired_values(self, t):
        # yd = []
        yd = np.zeros((self._settings['differential_order'] + 1), )

        a = self._settings['Amplitude']
        f = self._settings['Frequency']
        off = self._settings["Offset"]
        p = self._settings["Phase in degree"] * np.pi / 180

        for idx, val in enumerate(self.yd_sym):
            yd[idx] = val(t, a, f, off, p)
            # yd.append(val(t, a, f, off, p))

        return yd


class Setpoint(Trajectory):
    """
    Provides setpoints for every output component.

    If the output is not scalar, just add more entries to the list.
    By querying the differential order from the controller (if available) the
    required derivatives are given.

    Note:
        Keep in mind that while this class provides zeros for all derivatives
        of the desired value, they actually strive to infinity for :math:`t=0` .
    """

    public_settings = OrderedDict([("Setpoint", [0])])

    def __init__(self, settings):
        Trajectory.__init__(self, settings)
        self.yd = np.zeros((len(self._settings["Setpoint"]),
                            self._settings["differential_order"] + 1))

        for idx, val in enumerate(self._settings["Setpoint"]):
            self.yd[idx, 0] = val

    def _desired_values(self, t):
        return self.yd


class InterpolatorTrajectory(Trajectory):
    """
    Generic trajectory that interpolates between the given values

    This class basically is a wrapper for scipy's `interp1d`.

    Note:
        Values outside the given time domain will not be extrapolated but held
        at their last values.

    Parameters:
        kind (string): The type of interpolation to use, possible values are
        'linear', 'nearest', 'nearest-up', 'zero', 'slinear', 'quadratic', 'cubic', 'previous', or 'next'.
        x_data (array): Time steps to work on
        y_data (array): Corresponding values
    """
    public_settings = OrderedDict([
        ("kind", "zero"),
        ("x_data", [0, 1]),
        ("y_data", [0, 10]),
    ])

    def __init__(self, settings):
        super().__init__(settings)
        x_data = settings["x_data"]
        y_data = settings["y_data"]
        self._interp = interp1d(x=x_data,
                                y=y_data,
                                kind=settings["kind"],
                                bounds_error=False,
                                fill_value=(y_data[0], y_data[-1]),
                                )

    def _desired_values(self, t):
        return self._interp(t)


class Feedthrough(Feedforward):
    """
    A simple feedthrough that passes the reference trajectory to its output.
    """
    public_settings = OrderedDict([
        ("tick divider", 1),
    ])

    def __init__(self, settings):
        settings.update(input_order=0)
        Feedforward.__init__(self, settings)

    def _feedforward(self, time, trajectory_values):
        return trajectory_values


class LinearStateSpaceController(Controller):
    """
    A controller that is based on a state space model of a linear system.

    This controller needs a linear statespace model, just as the
    :py:class:`LinearStateSpaceModel` . The file provided in ``config file``
    should therefore contain a dict holding the entries: ``model``,
    ``op_inputs`` and ``op_outputs`` .

    If poles is given (differing from `None` ) the state-feedback will be
    computed using :py:func:`pymoskito.place_siso` .
    Furthermore, an appropriate prefilter is calculated, which establishes
    stationary attainment of the desired output values.

    Note:
        If a SIMO or MIMO system is given, the control_ package as well as the
        slycot_ package are needed the perform the pole placement.


    .. _control: https://github.com/python-control/python-control
    .. _slycot: https://github.com/python-control/Slycot
    """

    public_settings = OrderedDict([
        ("input source", "Model_State"),
        ("config file", None),
        ("poles", None),
    ])

    def __init__(self, settings):
        file = settings["config file"]
        assert os.path.isfile(file)

        with open(file, "rb") as f:
            data = pickle.load(f)

        if "system" not in data:
            raise ValueError("Config file lacks mandatory settings.")

        self.ss = data["system"]

        self.input_offset = data.get("op_inputs", None)
        self.output_offset = data.get("op_outputs", None)

        if self.input_offset is None:
            self.input_offset = np.zeros((self.ss.B.shape[1], ))
        if len(self.input_offset) != self.ss.B.shape[1]:
            raise ValueError("Provided input offset does not match input "
                             "dimensions.")

        if self.output_offset is None:
            self.output_offset = np.zeros((self.ss.C.shape[0], ))
        if len(self.output_offset) != self.ss.C.shape[0]:
            raise ValueError("Length of provided output offset does not match "
                             "output dimensions ({} != {}).".format(
                len(self.output_offset),
                self.ss.C.shape[0]
            ))

        # add specific "private" settings
        settings.update(input_order=0)
        settings.update(output_dim=self.ss.C.shape[0])
        settings.update(input_type=settings["input source"])
        super().__init__(settings)

        if settings.get("poles", None) is None:
            # pretty useless but hey why not.
            self.feedback = np.zeros((self.ss.B.shape[1], self.ss.A.shape[0]))
        else:
            if self.ss.B.shape[1] == 1:
                # save the control/slycot dependency
                self.feedback = place_siso(self.ss.A,
                                           self.ss.B,
                                           self.settings["poles"])
            else:
                import control
                self.feedback = control.place(self.ss.A,
                                              self.ss.B,
                                              self.settings["poles"])

        self.prefilter = calc_prefilter(self.ss.A, self.ss.B, self.ss.C,
                                        self.feedback)

    def _control(self, time, trajectory_values=None, feedforward_values=None,
                 input_values=None, **kwargs):
        return (-self.feedback @ input_values
                + self.prefilter @ (trajectory_values[:, 0] -
                                    self.output_offset)
                + self.input_offset)


class PIDController(Controller):
    """
    A simple scalar PID Controller with basic anti-windup

    This class will run a PID controller with the given settings for each
    component of the model output listed in 'input indexes'.
    """
    public_settings = OrderedDict([
        ("input indexes", [0]),
        ("Kp", 1),
        ("Ki", 1),
        ("Kd", 1),
        ("output limits", [-1, 1]),
        ("tick divider", 1)])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(input_order=0)  # no reference derivatives required
        settings.update(input_type="Model_Output")  # use the model output to compute the control error
        output_dim = len(self.public_settings["input indexes"])
        settings.update(output_dim=output_dim)
        Controller.__init__(self, settings)

        self.dt = settings["step size"]
        self.error_integral = np.zeros(output_dim)
        self.previous_error = np.zeros(output_dim)

    def _control(self, time, trajectory_values=None, feedforward_values=None, input_values=None, **kwargs):
        #  compute the control error
        y = np.asarray(input_values)[self._settings["input indexes"]]
        yd = np.atleast_2d(trajectory_values)[:, 0]  # ignore reference derivatives
        error = yd - y

        # compute the error integral
        integral = self.error_integral + error * self.dt

        # compute the error derivative
        differential = (error - self.previous_error) / self.dt

        # compute control outputs
        output = (self._settings["Kp"] * error
                          + self._settings["Ki"] * integral
                          + self._settings["Kd"] * differential)

        # saturate the outputs
        sat_output = np.clip(output, *self._settings["output limits"])

        # only integrate errors of unsaturated outputs
        unsat_idxs = output == sat_output
        self.error_integral[unsat_idxs] = integral[unsat_idxs]

        # store current error
        self.previous_error = error

        return sat_output


class AdditiveMixer(SignalMixer):
    """
    Signal Mixer that accumulates all input signals.

    Processing is done according to rules of numpy broadcasting.
    """
    public_settings = OrderedDict([("Input A", None),
                                   ("Input B", None)])

    def __init__(self, settings):
        settings.update([("input signals", [settings["Input A"],
                                            settings["Input B"]])])
        SignalMixer.__init__(self, settings)

    def _mix(self, signal_values):
        return sum(signal_values)


class ModelInputLimiter(Limiter):
    """
    ModelInputLimiter that limits the model input values.

    Settings:
        `Limits`: (List of) list(s) that hold (min, max) pairs for the
        corresponding input.
    """

    public_settings = OrderedDict([("Limits", [None, None])])

    def __init__(self, settings):
        settings.update([("input_signal", "ModelMixer")])
        Limiter.__init__(self, settings)
        self.limits = np.atleast_2d(settings["Limits"])

    def _limit(self, values):
        val = np.atleast_1d(values)
        out = np.zeros_like(val)
        for idx, v in enumerate(val):
            lim = self.limits[idx]
            if lim[0] is None and lim[1] is None:
                out[idx] = v
            else:
                out[idx] = np.clip(v, *lim)

        if len(values.shape) == 1:
            return out.flatten()
        else:
            return out


class DeadTimeSensor(Sensor):
    """
    Sensor that adds a measurement delay on chosen states
    """

    public_settings = OrderedDict([("states to delay", [0]),
                                   ("delay", 1)])

    def __init__(self, settings):
        settings.update([("input signal", "Model_State")])
        Sensor.__init__(self, settings)
        self._storage = None

    def _measure(self, value):
        if self._storage is None:
            # create storage with length "delay"
            # initial values are the first input
            self._storage = [value]*int(self._settings["delay"])

        # save current values
        measurement = value.copy()
        # add new measurement
        self._storage.append(value)

        # get delayed measurements
        delayed_measurement = self._storage.pop(0)

        # replace current values with delayed values, if it is chosen
        for i in self._settings["states to delay"]:
            measurement[i] = delayed_measurement[i]

        return measurement


class GaussianNoise(Disturbance):
    """
    Noise generator for gaussian noise
    """

    public_settings = OrderedDict([("sigma", 1),
                                   ("mean", 0)])

    def __init__(self, settings):
        settings.update([("input signal", "Model_Output")])
        Disturbance.__init__(self, settings)

    def _disturb(self, t, signal):
        return np.random.normal(self._settings['mean'],
                                self._settings['sigma'],
                                signal.shape)



# -*- coding: utf-8 -*-
import matplotlib as mpl

mpl.rcParams['text.usetex'] = True
mpl.rcParams['text.latex.unicode'] = True
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

from processing_gui import MetaProcessingModule
import settings as st
from tools import sort_tree


class eval_L1Norms_param_linePlot(MetaProcessingModule):
    """
    create diagrams for evaluation of integralError
    """
    name = 'eval_L1_over_param'

    def __init__(self):
        MetaProcessingModule.__init__(self)
        return

    def run(self, post_results):
        # create tree with relevant data
        source = sort_tree(post_results, ['modules', 'controller', 'type'])

        # get ideal regime name
        file_name = post_results[0]['name']

        # extract parameter that has been varied
        param = next((param for param in st.paramVariationListB \
                      if '_' + param + '(' in file_name), \
                     None)
        self._logger.info("assuming that {} has been varied".format(param))

        # create plot
        figs = [Figure(), Figure()]

        # plot for L1NormAbs
        axes = [figs[0].add_subplot(111), figs[1].add_subplot(111)]

        self.plotVariousController(source, axes[0],
                                   xPath=['modules', 'model', param],
                                   yPath=['metrics', 'L1NormAbs'],
                                   typ='line')

        if param == 'M':
            xlbl = r'$M \, \lbrack kg\rbrack$'
        elif param == 'J':
            xlbl = r'$M \, \lbrack kg\rbrack$'
        else:
            xlbl = ''

        self.plotSettings(axes[0],
                          titel='Fehlerintegral',
                          grid=True,
                          xlabel=xlbl,
                          ylabel=r'$E \, \lbrack ms\rbrack$')
        self.plotVariousController(source, axes[1],
                                   xPath=['modules', 'model', param],
                                   yPath=['metrics', 'L1NormITAE'],
                                   typ='line')
        self.plotSettings(axes[1],
                          titel='ITAE Fehler',
                          grid=True,
                          xlabel=xlbl,
                          ylabel=r'$E \, \lbrack ms^2\rbrack$')

        # extract controller names
        controller_names = [x[:-len('Controller')] for x in source.keys()]

        canvas = [FigureCanvas(figs[0]), FigureCanvas(figs[1])]

        # write output files
        file_name = self.name[len('eval_'):] \
                    + '_Controller_(' + ''.join(controller_names) + ')'

        names = ['_'.join([file_name, param, 'Abs']), '_'.join([file_name, param, 'ITAE'])]

        self.writeOutputFiles(names[0], figs[0])
        self.writeOutputFiles(names[1], figs[1])

        return [
            {'figure': canvas[0], 'name': names[0]},
            {'figure': canvas[1], 'name': names[1]},
        ]



# -*- coding: utf-8 -*-

#mpl.rcParams['text.usetex']=True
#mpl.rcParams['text.latex.unicode']=True
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

from processing_gui import MetaProcessingModule
from tools import sortTree


class eval_L1NormAbs_delta_t_linePlot(MetaProcessingModule):
    '''
    create diagrams for evaluation of L1NormAbs
    '''
    def __init__(self):
        MetaProcessingModule.__init__(self)
        return

    def run(self, postResults):
        #create tree with relevant datas
        source = sortTree(postResults, ['modules', 'controller', 'type'])
        
        #create plot
        fig = Figure()
        
        #plot for L1NormAbs
        axes = fig.add_subplot(211)
        self.plotVariousController(source, axes,\
                xPath=['modules','trajectory', 'delta t'],\
                yPath=['metrics','L1NormAbs'],\
                typ='line')
        self.plotSettings(axes,\
                titel=r'Fehlerintegral w(t) und y(t) \"uber $\Delta t$',\
                grid=True,\
                xlabel=r'$\Delta t \, \lbrack s\rbrack$',\
                ylabel=r'$E \, \lbrack ms\rbrack$',\
                )
                        
        #plot for time-difference
        axes1 = fig.add_subplot(212)
        self.plotVariousController(source, axes1,\
                xPath=['modules','trajectory', 'delta t'],\
                yPath=['metrics','t_diff'],\
                typ='line')
        self.plotSettings(axes1,\
                titel=r'\"Ubergangszeitfehler \"uber $\Delta t$',\
                grid=True,\
                xlabel=r'$\Delta t \, \lbrack s\rbrack$',\
                ylabel=r'$e_{t} \, \lbrack s\rbrack$',\
                )
        
        # spacing
        fig.subplots_adjust(wspace=0.5, hspace=0.5)
        
        #extract controllerNames
        controllerNames = [x[:-len('Controller')] for x in source.keys()]
        
        canvas = FigureCanvas(fig)
        #write output files
        fileName = self.name[len('eval_'):]\
                    + '_Controller_(' + ''.join(controllerNames) + ')'
        self.writeOutputFiles(fileName, fig)
        
        return [{'figure': canvas, 'name': self.name}]



import numpy as np
from collections import OrderedDict

import pymoskito as pm

from . import settings as st

class BallInTubeModel(pm.Model):
    """
    Implementation of the Ball in Tube System

    Collisions with the floor and the ceilings are modelled as a
    overdamped mass-spring-damper system.
    """
    public_settings = OrderedDict([('initial state', st.initial_state),
                                   ('d_B', st.d_B),
                                   ('d_R', st.d_R),
                                   ('tube_length', st.tube_length),
                                   ('k_L', st.k_L),
                                   ('k_V', st.k_V),
                                   ('k_s', st.k_s),
                                   ('d', st.d),
                                   ('T', st.T),
                                   ('m', st.m),
                                   ('g', st.g),
                                   ('K', st.K),
                                   ('D', st.D),
                                   ])

    def __init__(self, settings):
        # add specific "private" settings
        settings.update(state_count=4)
        settings.update(input_count=1)
        settings.update({"output_info": {
            0: {"Name": "ball elevation", "Unit": "m"},
        }})
        pm.Model.__init__(self, settings)

        # register events
        self.register_event(self.ball_event_lower)
        self.register_event(self.ball_event_upper)

        # shortcuts for readability
        self.d_B = self._settings['d_B']
        self.d_R = self._settings['d_R']
        self.tube_length = self._settings['tube_length']
        self.k_L = self._settings['k_L']
        self.k_V = self._settings['k_V']
        self.k_s = self._settings['k_s']
        self.d = self._settings['d']
        self.T = self._settings['T']
        self.m = self._settings['m']
        self.g = self._settings['g']
        self.K = self._settings['K']
        self.D = self._settings['D']
        self.A_B = np.pi*self.d_B**2/4
        self.A_R = np.pi*self.d_R**2/4
        self.A_Sp = self.A_R - self.A_B

    def state_function(self, t, x, args):
        """
        Calculations of system state changes

        If the ball should get in contact with either the tube flor or ceiling,
        an extra spring with spring stiffness K and a damping D is added to simulate
        the contact.

        :param x: state
        :param t: time
        :type args: system input u
        """
        # definitional
        x1 = x[0]
        x2 = x[1]
        x3 = x[2]
        x4 = x[3]
        u = args[0].squeeze()

        dx1 = x2
        dx2 = -x1/self.T**2 - 2*self.d*x2/self.T + self.k_s*u*12/(255*self.T**2)
        dx3 = x4

        dx4 = (self.k_L * ((self.k_V * x1 - self.A_B * x4) / self.A_Sp) ** 2) / self.m - self.g
        if x3 < 0:
            # contact with the floor
            dx4 += -self.K*x3/self.m - (self.D*x4)/self.m
        if x3 + self.d_B > self.tube_length:
            # contact with the ceiling
            dx4 += -self.K * (x3 + self.d_B - self.tube_length)/self.m - self.D*x4/self.m

        return np.array([dx1, dx2, dx3, dx4])

    def ball_event_lower(self, t, x):
        return x[2]

    def ball_event_upper(self, t, x):
        return x[2] + self.d_B - self.tube_length

    def calc_output(self, input_vector):
        """
        return ball position as output
        :param input_vector: input values
        :return: ball position
        """
        return input_vector[2]


pm.register_simulation_module(pm.Model, BallInTubeModel)



# -*- coding: utf-8 -*-
import os

# mpl.rcParams['text.usetex'] = True
# mpl.rcParams['text.latex.unicode']=True
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

from processing_gui import MetaProcessingModule


class eval_itae(MetaProcessingModule):
    '''
    create diagrams for evaluation of itea metric
    '''

    line_color = '#aaaaaa'
    line_style = '-'
    font_size = 20
    epsPercent = 2.5
    spacing = 0.01
    counter = 0

    def __init__(self):
        MetaProcessingModule.__init__(self)
        return

    def run(self, postResults):
        # aquire data
        errors = [elem['error_L1Norm'] for elem in postResults]
        poles = [elem['modules']['observer']['poles'][0] for elem in postResults]

        # create plot
        fig = Figure()
        axes = fig.add_subplot(111)
        axes.set_title(r'\textbf{L1 Kriterium}')
        axes.plot(poles, errors, c='k')
        # axes.set_xlim(left=0, right=t[-1])
        # axes.set_xlabel(r'\textit{Zeit [s]}')
        # axes.set_ylabel(r'\textit{Ballposition r(t) [m]}')

        # write results
        filePath = os.path.join(os.path.pardir, 'results', 'metaprocessing', 'C')
        if not os.path.isdir(filePath):
            os.makedirs(filePath)

        fileName = os.path.join(filePath, 'L1_observer-plot')
        canvas = FigureCanvas(fig)
        fig.savefig(fileName + '.svg')
        fig.savefig(fileName + '.png')

        results = [{'figure': canvas, 'name': 'L1_observer-plot'}, \
                   ]

        return results



# -*- coding: utf-8 -*-

import matplotlib as mpl

mpl.rcParams['text.usetex'] = True
mpl.rcParams['text.latex.unicode'] = True
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

from processing_gui import MetaProcessingModule
from tools import sortTree


class eval_L1NormAbs_sigma_linePlot(MetaProcessingModule):
    '''
    create diagrams for evaluation of L1NormAbs
    '''

    def __init__(self):
        MetaProcessingModule.__init__(self)
        return

    def run(self, postResults):
        # create tree with relevant datas
        source = sortTree(postResults, ['modules', 'controller', 'type'])

        # create plot
        fig = Figure()
        fig.subplots_adjust(wspace=0.5, hspace=0.25)

        # plot for L1NormAbs
        axes = fig.add_subplot(111)

        self.plotVariousController(source, axes, \
                                   xPath=['modules', 'disturbance', 'sigma'], \
                                   yPath=['metrics', 'L1NormAbs'], \
                                   typ='line')
        self.plotSettings(axes, \
                          titel=r'Fehlerintegral w(t) und y(t) \"uber Sigma', \
                          grid=True, \
                          xlabel=r'$\sigma \, \lbrack m\rbrack$', \
                          ylabel=r'$E \, \lbrack ms \rbrack$', \
                          )

        # extract controllerNames
        controllerNames = [x[:-len('Controller')] for x in source.keys()]

        canvas = FigureCanvas(fig)
        # write output files
        fileName = self.name[len('eval_'):] \
                   + '_Controller_(' + ''.join(controllerNames) + ')'
        self.writeOutputFiles(fileName, fig)

        return [{'figure': canvas, 'name': self.name}]



# -*- coding: utf-8 -*-
import os

import matplotlib as mpl
import settings as st

mpl.rcParams['text.usetex'] = True
mpl.rcParams['text.latex.unicode'] = True
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

from processing_gui import MetaProcessingModule


class eval_L1perFA05(MetaProcessingModule):
    '''
    create diagrams for evaluation of itea metric
    '''

    line_color = '#aaaaaa'
    line_style = '-'
    font_size = 20
    epsPercent = 2.5
    spacing = 0.01
    counter = 0

    def __init__(self):
        MetaProcessingModule.__init__(self)
        return

    def sortLists(self, val):
        val[1] = [x for (y, x) in sorted(zip(val[0], val[1]))]
        val[0] = sorted(val[0])
        return val

    def run(self, postResults):
        controllerDict = {'FController': [[], []], \
                          'GController': [[], []], \
                          'JController': [[], []], \
                          'LSSController': [[], []], \
                          'PIFeedbackController': [[], []]}

        # TODO: levels per input füllen
        level1 = 'modules'
        level2 = 'trajectory'
        level3 = 'Frequency'
        xLabel = 'f [Hz]'
        yLabel = 'E [m^2]'

        for elem in postResults:
            controllerDict[elem['modules']['controller']['type']][0].append(elem[level1][level2][level3])
            controllerDict[elem['modules']['controller']['type']][1].append(elem['metrics']['L1NormAbs'])

        fig = Figure()
        axes = fig.add_subplot(1, 1, 1)

        xMax = 0
        leg = []
        for elem in controllerDict:
            controllerDict[elem] = self.sortLists(controllerDict[elem])
            axes.plot(controllerDict[elem][0], controllerDict[elem][1], 'o-', \
                      c=st.color_cycle[elem])
            leg.append(elem)
            if controllerDict[elem][0]:
                if xMax < controllerDict[elem][0][-1]:
                    xMax = controllerDict[elem][0][-1]
        axes.legend(leg, loc=0)
        axes.set_xlim(left=0.1, right=xMax)
        #        axes.set_ylim(top=6.0, bottom=3)
        axes.set_xlabel(r'$' + xLabel + '$', size=st.label_size)
        axes.set_ylabel(r'$' + yLabel + '$', size=st.label_size)
        axes.set_title(r'Fehlerintegral \"uber Frequenz', size=st.label_size)
        axes.grid(color='#ababab', linestyle='--')

        # write results
        filePath = os.path.join(os.path.pardir, 'results', 'metaprocessing', 'A2')
        if not os.path.isdir(filePath):
            os.makedirs(filePath)

        metaName = 'L1-plotFA05'
        fileName = os.path.join(filePath, metaName)
        canvas = FigureCanvas(fig)
        fig.savefig(fileName + '.svg')
        fig.savefig(fileName + '.png')
        fig.savefig(fileName + '.pdf')

        results = [{'figure': canvas, 'name': metaName}, \
                   ]

        return results



# -*- coding: utf-8 -*-
from tools import getSubValue

# mpl.rcParams['text.usetex']=True
# mpl.rcParams['text.latex.unicode']=True
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

from processing_gui import MetaProcessingModule
from tools import sortTree
from tools import getSubValue


class eval_L1NormITAE_poles_linePlot(MetaProcessingModule):
    """
    create diagrams, which plot L1NormITAE over poles
    """

    def __init__(self):
        MetaProcessingModule.__init__(self)
        return

    def run(self, post_results):
        # create tree with all relevant data
        source = sortTree(post_results, ['modules', 'controller', 'type'])

        # Get the Poles for the Minimum 
        for controller in source:
            x_path = ['modules', 'controller', 'poles']
            y_path = ['metrics', 'L1NormITAE']
            x_list = getSubValue(source[controller], x_path)
            y_list = getSubValue(source[controller], y_path)

            x_list[:] = [x for x, y in zip(x_list, y_list) if y]
            y_list[:] = [i for i in y_list if i]

            self._logger.info("processing controller '{}'".format(controller))
            self._logger.info("min ITAE {}".format(min(y_list)))
            self._logger.info("opt poles {}".format(x_list[y_list.index(min(y_list))][0]))

        # create plot
        fig = Figure()
        #        fig.subplots_adjust(wspace=20, hspace=10, h_pad=50)
        fig.subplots_adjust(wspace=0.6, hspace=0.6)

        # plot for L1NormITAE over poles
        axes = fig.add_subplot(111)
        self.plotVariousController(source, axes,
                                   xPath=['modules', 'controller', 'poles'],
                                   yPath=['metrics', 'L1NormITAE'],
                                   typ='line',
                                   xIndex=0)
        self.plotSettings(axes,
                          titel=r'Fehlerintegral ITAE \"uber Polplatzierung',
                          grid=True,
                          xlabel=r'$Poles \, \lbrack s\rbrack$',
                          ylabel=r'$E \, \lbrack ms^{2} \rbrack$',
                          )

        # error minimum
        for controllerName in source.keys():
            error_list = getSubValue(source[controllerName], ['metrics', 'L1NormITAE'])
            error_min = min(x for x in error_list if x is not None)
            error_min_index = error_list.index(error_min)
            poles = getSubValue(source[controllerName], ['modules', 'controller', 'poles'])[error_min_index][0]
            self._logger.info("minimum error of {} for {} with poles at {}".format(error_min, controllerName, poles))

        # extract controllerNames
        controller_names = [x[:-len('Controller')] for x in source.keys()]

        canvas = FigureCanvas(fig)
        # write output files
        file_name = self.name[len('eval_'):] \
                    + '_Controller_(' + ''.join(controller_names) + ')'
        self.writeOutputFiles(file_name, fig)

        return [{'figure': canvas, 'name': self.name}]



# -*- coding: utf-8 -*-
from importlib import metadata
import logging
from operator import itemgetter
import os
import pickle
import time
import webbrowser
import yaml

import numpy as np
from scipy.interpolate import interp1d

from PyQt5.QtCore import (pyqtSignal, pyqtSlot, Qt, QTimer, QSize, QSettings,
                          QCoreApplication, QModelIndex, QFile, QTextStream)
from PyQt5.QtGui import QIcon, QKeySequence, QColor, QPalette
from PyQt5.QtWidgets import (
    QApplication, QMainWindow,
    QWidget, QAction,
    QSlider, QLabel, QFrame,
    QPushButton, QComboBox,
    QListWidget, QListWidgetItem, QTreeView, QTreeWidget, QTreeWidgetItem,
    QAbstractItemView,
    QToolBar, QStatusBar, QProgressBar,
    QTextEdit, QFileDialog, QInputDialog,
    QVBoxLayout, QHBoxLayout, QMessageBox,
)

# pyqtgraph
import pyqtgraph as pg
from pyqtgraph.dockarea import DockArea

# vtk
vtk_error_msg = ""
try:
    import vtk

    from vtk import vtkRenderer
    from vtk import qt
    # import patched class that fixes scroll problem
    from .visualization import QVTKRenderWindowInteractor

    vtk_available = True
except ImportError as e:
    vtk_available = False
    vtk_error_msg = e
    vtkRenderer = None
    QVTKRenderWindowInteractor = None

# pymoskito
from .registry import get_registered_visualizers
from .simulation_interface import SimulatorInteractor, SimulatorView
from .visualization import MplVisualizer, VtkVisualizer, DummyVisualizer
from .processing_gui import PostProcessor
from .resources.themes import darkPalette, lightPalette, DarkModeStyle, LightModeStyle
from .tools import (
    get_resource, PlainTextLogger, Exporter, create_button_from_action
)

__all__ = ["SimulationGui", "run"]


def run(regimes=None, show=True, execute=False, exit_after=False):
    """
    Helper function to launch the PyMoskito GUI

    Args:
        regimes (String): Path to a regime file to load
        show (bool): If true, show the GUI otherwise operate in headless mode
        execute (bool): If true, simulate all loaded regimes
        exit_after (bool): If true, exit after all regimes are simulated
    """
    app = QApplication([])
    prog = SimulationGui()
    if show:
        prog.show()
    if regimes is not None:
        prog.load_regimes_from_file(regimes)
    if exit_after:
        prog.actExitOnBatchCompletion.setChecked(True)
    if execute:
        prog.start_regime_execution()
    app.exec_()


class SimulationGui(QMainWindow):
    """
    The graphical user interface of PyMoskito
    """
    # TODO enable closing plot docks by right-clicking their name
    TABLEAU_COLORS = (
        ('blue', '#1f77b4'),
        ('orange', '#ff7f0e'),
        ('green', '#2ca02c'),
        ('red', '#d62728'),
        ('purple', '#9467bd'),
        ('brown', '#8c564b'),
        ('pink', '#e377c2'),
        ('gray', '#7f7f7f'),
        ('olive', '#bcbd22'),
        ('cyan', '#17becf'),
    )

    runSimulation = pyqtSignal()
    stopSimulation = pyqtSignal()
    playbackTimeChanged = pyqtSignal()
    regimeFinished = pyqtSignal()
    finishedRegimeBatch = pyqtSignal(bool)

    def __init__(self):
        # constructor of the base class
        QMainWindow.__init__(self)

        QCoreApplication.setOrganizationName("PyMoskito")
        QCoreApplication.setOrganizationDomain("https://github.com/cklb/pymoskito")
        QCoreApplication.setApplicationVersion(metadata.version("pymoskito"))
        QCoreApplication.setApplicationName(globals()["__package__"])

        # load settings
        self._settings = QSettings()
        self._init_settings()

        # initialize logger
        self._logger = logging.getLogger(self.__class__.__name__)

        # Create Simulation Backend
        self.guiProgress = None
        self.cmdProgress = None
        self.sim = SimulatorInteractor(self)
        self.runSimulation.connect(self.sim.run_simulation)
        self.stopSimulation.connect(self.sim.stop_simulation)
        self.sim.simulation_finalized.connect(self.new_simulation_data)
        self.currentDataset = None
        self.inputInterpolator = None
        self.stateInterpolator = None

        # sim setup viewer
        self.targetView = SimulatorView(self)
        self.targetView.setModel(self.sim.target_model)
        self.targetView.expanded.connect(self.target_view_changed)
        self.targetView.collapsed.connect(self.target_view_changed)

        # sim results viewer
        self.result_view = QTreeView()

        # icons
        QIcon.setThemeSearchPaths(QIcon.themeSearchPaths() + [get_resource(".")])
        self.setTheme()
        icon_size = QSize(25, 25)

        # the docking area allows to rearrange the user interface at runtime
        self.area = pg.dockarea.DockArea()

        # Window properties
        self.setCentralWidget(self.area)
        self.resize(1000, 700)
        self.setWindowTitle("PyMoskito")
        res_path = get_resource("mosquito")
        icon = QIcon(res_path)
        self.setWindowIcon(icon)

        # create docks
        self.propertyDock = pg.dockarea.Dock("Properties")
        self.animationDock = pg.dockarea.Dock("Animation")
        self.regimeDock = pg.dockarea.Dock("Regimes")
        self.lastSimDock = pg.dockarea.Dock("Previous Simulations")
        self.dataDock = pg.dockarea.Dock("Data & Plots")
        self.logDock = pg.dockarea.Dock("Log")

        # arrange docks
        self.area.addDock(self.animationDock, "right")
        self.area.addDock(self.lastSimDock, "left", self.animationDock)
        self.area.addDock(self.propertyDock, "bottom", self.lastSimDock)
        self.area.addDock(self.dataDock, "bottom", self.propertyDock)
        self.area.addDock(self.logDock, "bottom", self.dataDock)
        self.area.addDock(self.regimeDock, "left", self.lastSimDock)
        self.non_plotting_docks = list(self.area.findAll()[1].keys())
        self.standardDockState = self.area.saveState()
        if state := self._settings.value("view/dock_state", type=dict):
            self.area.restoreState(state, missing="ignore")

        # add widgets to the docks
        self.propertyDock.addWidget(self.targetView)

        # setup the visualizer
        self.visuComboBox = QComboBox()
        self.visualizer = None
        self._init_visualizer()

        # regime window
        self.regime_list = QListWidget(self)
        self.regime_list.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.regimeDock.addWidget(self.regime_list)
        self.regime_list.itemDoubleClicked.connect(self.regime_dclicked)
        self._regimes = []
        self.regime_file_name = ""

        self.actDeleteRegimes = QAction(self.regime_list)
        self.actDeleteRegimes.setText("&Delete Selected Regimes")
        # TODO shortcut works always, not only with focus on the regime list
        # self.actDeleteRegimes.setShortcutContext(Qt.WindowShortcut)
        self.actDeleteRegimes.setShortcut(QKeySequence(Qt.Key_Delete))
        self.actDeleteRegimes.triggered.connect(self.remove_regime_items)

        self.actSave = QAction(self)
        self.actSave.setText('Save Results As')
        self.actSave.setIcon(QIcon.fromTheme("save"))
        self.actSave.setDisabled(True)
        self.actSave.setShortcut(QKeySequence.Save)
        self.actSave.triggered.connect(self.export_simulation_data)

        self.actLoadRegimes = QAction(self)
        self.actLoadRegimes.setText("Load Regimes from File")
        self.actLoadRegimes.setIcon(QIcon.fromTheme("load"))
        self.actLoadRegimes.setDisabled(False)
        self.actLoadRegimes.setShortcut(QKeySequence.Open)
        self.actLoadRegimes.triggered.connect(self.load_regime_dialog)

        self.actExitOnBatchCompletion = QAction(self)
        self.actExitOnBatchCompletion.setText("&Exit On Batch Completion")
        self.actExitOnBatchCompletion.setCheckable(True)
        self.actExitOnBatchCompletion.setChecked(
            self._settings.value("control/exit_on_batch_completion", type=bool)
        )
        self.actExitOnBatchCompletion.changed.connect(
            self.update_exit_on_batch_completion_setting)

        # regime management
        self.runningBatch = False
        self._current_regime_index = None
        self._current_regime_name = None

        self.regimeFinished.connect(self.run_next_regime)
        self.finishedRegimeBatch.connect(self.regime_batch_finished)

        # last sim window
        self.lastSimList = QListWidget(self)
        self.lastSimList.setFocusPolicy(Qt.NoFocus)
        self.lastSimDock.addWidget(self.lastSimList)
        self._lastSimulations = []
        self.lastSimList.itemDoubleClicked.connect(self.load_last_sim)

        # data window
        self.dataWidget = QWidget()
        self.dataLayout = QHBoxLayout()

        self.dataPointListWidget = QListWidget()
        self.dataPointListLayout = QVBoxLayout()
        self.dataPointListWidget.setLayout(self.dataPointListLayout)
        self.dataPointListWidget.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.dataLayout.addWidget(self.dataPointListWidget)

        self.dataPointManipulationLayout = QVBoxLayout()
        self.dataPointManipulationLayout.addStretch()
        self.dataPointManipulationLayout.setSpacing(5)
        self.dataPointManipulationWidget = QWidget()
        self.dataPointManipulationWidget.setLayout(self.dataPointManipulationLayout)
        self.dataLayout.addWidget(self.dataPointManipulationWidget)

        self.dataPointLabel = QLabel("Data", self)
        self.dataPointLabel.setAlignment(Qt.AlignCenter)
        self.dataPointManipulationLayout.addWidget(self.dataPointLabel)

        self.actAddDataPointToPlot = QAction(self)
        self.actAddDataPointToPlot.setText("&Add selected data to plot")
        self.actAddDataPointToPlot.setIcon(QIcon.fromTheme("add"))
        self.actAddDataPointToPlot.setToolTip(
            "Add the selected data set from the left to the selected plot "
            "on the right.")
        self.actAddDataPointToPlot.triggered.connect(self.add_data_point_to_tree)
        self.dataPointManipulationLayout.addWidget(
            create_button_from_action(self.actAddDataPointToPlot))

        self.actExportDataPoint = QAction(self)
        self.actExportDataPoint.setText("&Export selected data")
        self.actExportDataPoint.setIcon(QIcon.fromTheme("export"))
        self.actExportDataPoint.setToolTip(
            "Export the selected data set from the left to a csv or png file."
        )
        self.actExportDataPoint.triggered.connect(self.export_data_points_from_tree)
        self.dataPointManipulationLayout.addWidget(
            create_button_from_action(self.actExportDataPoint))

        self.actRemoveDataPointFromPlot = QAction(self)
        self.actRemoveDataPointFromPlot.setText("&Remove selected data from plot")
        self.actRemoveDataPointFromPlot.setIcon(QIcon.fromTheme("delete"))
        self.actRemoveDataPointFromPlot.setToolTip(
            "Remove the selected data set from the plot on the right."
        )
        self.actRemoveDataPointFromPlot.triggered.connect(
            self.remove_data_point_from_tree)
        self.dataPointManipulationLayout.addWidget(
            create_button_from_action(self.actRemoveDataPointFromPlot))

        self.plotLabel = QLabel("Plots", self)
        self.plotLabel.setAlignment(Qt.AlignCenter)
        self.dataPointManipulationLayout.addWidget(self.plotLabel)

        self.actCreatePlot = QAction(self)
        self.actCreatePlot.setIcon(QIcon.fromTheme("add"))
        self.actCreatePlot.setText("&Create a new plot")
        self.actCreatePlot.setToolTip("Create a new plot.")
        self.actCreatePlot.triggered.connect(self.add_plot_tree_item)
        self.dataPointManipulationLayout.addWidget(
            create_button_from_action(self.actCreatePlot))

        self.actShowPlot = QAction(self)
        self.actShowPlot.setText("&Show the selected plot")
        self.actShowPlot.setIcon(QIcon.fromTheme("view"))
        self.actShowPlot.setToolTip(
            "Show the selected plot (double-click on name)."
        )
        self.actShowPlot.triggered.connect(self.show_selected_plot_tree_item)
        self.dataPointManipulationLayout.addWidget(
            create_button_from_action(self.actShowPlot))

        self.actDeletePlot = QAction(self)
        self.actDeletePlot.setText("&Delete selected plot")
        self.actDeletePlot.setIcon(QIcon.fromTheme("delete"))
        self.actDeletePlot.setToolTip(
            "Delete the selected plot."
        )
        self.actDeletePlot.triggered.connect(self.remove_selected_plot_tree_items)
        self.dataPointManipulationLayout.addWidget(
            create_button_from_action(self.actDeletePlot))
        self.dataPointManipulationLayout.addStretch()

        self.dataPointTreeWidget = QTreeWidget()
        self.dataPointTreeWidget.setHeaderLabels(["PlotTitle", "DataPoint"])
        # self.dataPointTreeWidget.setSelectionMode(QAbstractItemView.MultiSelection)
        self.dataPointTreeWidget.itemDoubleClicked.connect(self.plot_vector_double_clicked)
        self.dataPointTreeWidget.setExpandsOnDoubleClick(0)
        self.dataPointTreeLayout = QVBoxLayout()

        self.dataPointTreeWidget.setLayout(self.dataPointTreeLayout)
        self.dataLayout.addWidget(self.dataPointTreeWidget)

        self.dataWidget.setLayout(self.dataLayout)
        self.dataDock.addWidget(self.dataWidget)

        # actions for simulation control
        self.actSimulateCurrent = QAction(self)
        self.actSimulateCurrent.setText("&Simulate Current Regime")
        self.actSimulateCurrent.setIcon(QIcon.fromTheme("simulate"))
        self.actSimulateCurrent.setShortcut(QKeySequence("F5"))
        self.actSimulateCurrent.triggered.connect(self.start_simulation)

        self.actSimulateAll = QAction(self)
        self.actSimulateAll.setText("Simulate &All Regimes")
        self.actSimulateAll.setIcon(QIcon.fromTheme("execute_regimes"))
        self.actSimulateAll.setShortcut(QKeySequence("F6"))
        self.actSimulateAll.setDisabled(True)
        self.actSimulateAll.triggered.connect(self.start_regime_execution)

        # actions for animation control
        self.actAutoPlay = QAction(self)
        self.actAutoPlay.setText("&Autoplay Simulation")
        self.actAutoPlay.setCheckable(True)
        self.actAutoPlay.setChecked(
            self._settings.value("control/autoplay_animation", type=bool)
        )
        self.actAutoPlay.changed.connect(self.update_autoplay_setting)

        self.actPlayPause = QAction(self)
        self.actPlayPause.setText("Play Animation")
        self.actPlayPause.setIcon(QIcon.fromTheme("play"))
        self.actPlayPause.setDisabled(True)
        self.actPlayPause.setShortcut(QKeySequence(Qt.Key_Space))
        self.actPlayPause.triggered.connect(self.start_animation)

        self.actStop = QAction(self)
        self.actStop.setText("Stop")
        self.actStop.setIcon(QIcon.fromTheme("stop"))
        self.actStop.setDisabled(True)
        self.actStop.triggered.connect(self.stop_animation)

        self.actSlow = QAction(self)
        self.actSlow.setText("Slowest")
        self.actSlow.setIcon(QIcon.fromTheme("slow"))
        self.actSlow.setDisabled(False)
        self.actSlow.triggered.connect(self.set_slowest_playback_speed)

        self.actFast = QAction(self)
        self.actFast.setText("Fastest")
        self.actFast.setIcon(QIcon.fromTheme("fast"))
        self.actFast.setDisabled(False)
        self.actFast.triggered.connect(self.set_fastest_playback_speed)

        self.speedControl = QSlider(Qt.Horizontal, self)
        self.speedControl.setMaximumSize(200, 25)
        self.speedControl.setTickPosition(QSlider.TicksBothSides)
        self.speedControl.setDisabled(False)
        self.speedControl.setMinimum(0)
        self.speedControl.setMaximum(12)
        self.speedControl.setValue(6)
        self.speedControl.setTickInterval(6)
        self.speedControl.setSingleStep(2)
        self.speedControl.setPageStep(3)
        self.speedControl.valueChanged.connect(self.update_playback_speed)

        self.timeSlider = QSlider(Qt.Horizontal, self)
        self.timeSlider.setMinimum(0)
        self.timeSliderRange = 1000
        self.timeSlider.setMaximum(self.timeSliderRange)
        self.timeSlider.setTickInterval(1)
        self.timeSlider.setTracking(True)
        self.timeSlider.setDisabled(True)
        self.timeSlider.valueChanged.connect(self.update_playback_time)

        self.playbackTime = 0
        self.playbackGain = 1
        self.currentStepSize = 0
        self.currentStartTime = 0
        self.currentEndTime = 0
        self.playbackTimer = QTimer()
        self.playbackTimer.timeout.connect(self.increment_playback_time)
        self.playbackTimeChanged.connect(self.update_gui)
        self.playbackTimeout = 33  # in [ms] -> 30 fps

        self.actResetCamera = QAction(self)
        self.actResetCamera.setText("Reset Camera")
        self.actResetCamera.setIcon(QIcon.fromTheme("video_camera"))
        self.actResetCamera.setDisabled(True)
        self.actResetCamera.setEnabled(self.visualizer.can_reset_view)
        self.actResetCamera.triggered.connect(self.reset_camera_clicked)

        # postprocessing
        self.actPostprocessing = QAction(self)
        self.actPostprocessing.setText("Launch Postprocessor")
        self.actPostprocessing.setIcon(QIcon.fromTheme("processing"))
        self.actPostprocessing.setDisabled(False)
        self.actPostprocessing.triggered.connect(self.postprocessing_clicked)
        self.actPostprocessing.setShortcut(QKeySequence("F7"))
        self.postprocessor = None

        # theming
        self.actToggleTheme = QAction(self)
        self.actToggleTheme.setText("Toggle Theme")
        self.actToggleTheme.setIcon(QIcon.fromTheme("theme"))
        self.actToggleTheme.setDisabled(False)
        self.actToggleTheme.triggered.connect(self.toggleTheme)

        # toolbar
        self.toolbarSim = QToolBar("Simulation")
        self.toolbarSim.setContextMenuPolicy(Qt.PreventContextMenu)
        self.toolbarSim.setMovable(False)
        self.toolbarSim.setIconSize(icon_size)
        self.addToolBar(self.toolbarSim)
        self.toolbarSim.addAction(self.actLoadRegimes)
        self.toolbarSim.addAction(self.actSave)
        self.toolbarSim.addSeparator()
        self.toolbarSim.addAction(self.actSimulateCurrent)
        self.toolbarSim.addAction(self.actSimulateAll)
        self.toolbarSim.addSeparator()
        self.toolbarSim.addAction(self.actPlayPause)
        self.toolbarSim.addAction(self.actStop)
        self.toolbarSim.addWidget(self.timeSlider)
        self.toolbarSim.addSeparator()
        self.toolbarSim.addAction(self.actSlow)
        self.toolbarSim.addWidget(self.speedControl)
        self.toolbarSim.addAction(self.actFast)
        self.toolbarSim.addSeparator()
        self.toolbarSim.addAction(self.actResetCamera)
        self.toolbarSim.addWidget(self.visuComboBox)
        self.toolbarSim.addSeparator()
        self.toolbarSim.addAction(self.actPostprocessing)
        self.toolbarSim.addAction(self.actToggleTheme)

        # log dock
        self.logBox = QTextEdit(self)
        self.logBox.setReadOnly(True)
        self.logBox.setLineWrapMode(QTextEdit.NoWrap)
        self.logBox.ensureCursorVisible()
        self.logDock.addWidget(self.logBox)

        # init logger for logging box
        self.textLogger = PlainTextLogger(self._settings,
                                          logging.INFO)
        self.textLogger.set_target_cb(self.logBox)
        logging.getLogger().addHandler(self.textLogger)

        # menu bar
        fileMenu = self.menuBar().addMenu("&File")
        fileMenu.addAction(self.actLoadRegimes)
        fileMenu.addAction(self.actSave)
        fileMenu.addAction("&Quit", self.close, QKeySequence(Qt.CTRL + Qt.Key_W))

        editMenu = self.menuBar().addMenu("&Edit")
        editMenu.addAction(self.actDeleteRegimes)

        self.viewMenu = self.menuBar().addMenu('&View')
        self.actLoadStandardState = QAction('&Restore Default View')
        self.viewMenu.addAction(self.actLoadStandardState)
        self.actLoadStandardState.triggered.connect(self.loadStandardDockState)
        self.actShowCoords = QAction("&Show Coordinates", self)
        self.actShowCoords.setCheckable(True)
        self.actShowCoords.setChecked(
            self._settings.value("view/show_coordinates", type=bool)
        )
        self.viewMenu.addAction(self.actShowCoords)
        self.actShowCoords.changed.connect(self.update_show_coords_setting)

        simMenu = self.menuBar().addMenu("&Simulation")
        simMenu.addAction(self.actSimulateCurrent)
        simMenu.addAction(self.actSimulateAll)
        simMenu.addAction(self.actExitOnBatchCompletion)
        simMenu.addAction(self.actPostprocessing)

        animMenu = self.menuBar().addMenu("&Animation")
        animMenu.addAction(self.actPlayPause)
        animMenu.addAction("&Increase Playback Speed",
                           self.increment_playback_speed,
                           QKeySequence(Qt.CTRL + Qt.Key_Plus))
        animMenu.addAction("&Decrease Playback Speed",
                           self.decrement_playback_speed,
                           QKeySequence(Qt.CTRL + Qt.Key_Minus))
        animMenu.addAction("&Reset Playback Speed",
                           self.reset_playback_speed,
                           QKeySequence(Qt.CTRL + Qt.Key_0))
        animMenu.addAction(self.actAutoPlay)
        animMenu.addAction(self.actResetCamera)

        plotMenu = self.menuBar().addMenu("&Plotting")
        plotMenu.addAction(self.actAddDataPointToPlot)
        plotMenu.addAction(self.actExportDataPoint)
        plotMenu.addAction(self.actRemoveDataPointFromPlot)
        plotMenu.addSeparator()
        plotMenu.addAction(self.actCreatePlot)
        plotMenu.addAction(self.actShowPlot)
        plotMenu.addAction(self.actDeletePlot)
        plotMenu.addSeparator()
        self.actAskPlotTitle = QAction("Ask for plot title", self)
        self.actAskPlotTitle.setCheckable(True)
        self.actAskPlotTitle.setChecked(
            self._settings.value("plot/ask_for_plot_title", type=bool)
        )
        plotMenu.addAction(self.actAskPlotTitle)
        self.actAskPlotTitle.toggled.connect(
            lambda val: self._settings.setValue("plot/ask_for_plot_title", val))

        helpMenu = self.menuBar().addMenu("&Help")
        helpMenu.addAction("&Online Documentation", self.show_online_docs)
        helpMenu.addAction("&About", self.show_info)

        # status bar
        self.status = QStatusBar(self)
        self.setStatusBar(self.status)
        self.statusLabel = QLabel("Idle")
        self.statusBar().addPermanentWidget(self.statusLabel)
        self.timeLabel = QLabel("t=0.0")
        self.statusBar().addPermanentWidget(self.timeLabel)
        self.coordLabel = QLabel("x=0.0 y=0.0")
        self.statusBar().addPermanentWidget(self.coordLabel)

        self._logger.info("Simulation GUI is up and running.")

    def _init_visualizer(self):
        """ Initialize the visualizer combobox """
        self.visuComboBox.currentTextChanged.connect(self._visualizer_changed)
        self.visuComboBox.addItem("None")

        if not vtk_available:
            self._logger.warning(
                "loading vtk failed with:{}".format(vtk_error_msg))

        # get all registered visualizers
        available_vis = get_registered_visualizers()

        # fill the combo box
        for cls, name in available_vis:
            self.visuComboBox.addItem(name)

        # instantiate the first visualizer or None
        if self.visuComboBox.count() > 0:
            self.visuComboBox.setCurrentIndex(1)
        else:
            self.visuComboBox.setCurrentIndex(0)

    @pyqtSlot(str)
    def _visualizer_changed(self, name):
        """ Slot to update the visualizer """
        self._logger.info(
            "selecting visualizer '{}'".format(name))
        available_vis = get_registered_visualizers()
        cls = next((_c for _c, _n in available_vis if _n == name),
                   DummyVisualizer)
        self._update_visualizer(cls)

    def _update_visualizer(self, visu_cls):
        """ Update the current visualizer """
        if self.animationDock.widgets:
            self.animationDock.layout.removeWidget(
                self.animationDock.widgets[0])
            self.animationDock.widgets[0].deleteLater()
            del self.animationDock.widgets[0]

        self.animationLayout = QVBoxLayout()
        if visu_cls is DummyVisualizer:
            self.visualizer = DummyVisualizer()
            # self.animationWidget = QWidget()
            # self.animationDock.addWidget(self.animationWidget, row=0)
        elif issubclass(visu_cls, MplVisualizer):
            self.animationWidget = QWidget()
            self.visualizer = visu_cls(self.animationWidget,
                                       self.animationLayout)
            self.animationDock.addWidget(self.animationWidget, row=0)
        elif issubclass(visu_cls, VtkVisualizer):
            if vtk_available:
                # vtk window
                self.animationFrame = QFrame()
                self.vtkWidget = QVTKRenderWindowInteractor(
                    self.animationFrame)
                self.animationLayout.addWidget(self.vtkWidget)
                self.animationFrame.setLayout(self.animationLayout)
                self.animationDock.addWidget(self.animationFrame, row=0)
                self.vtk_renderer = vtkRenderer()
                self.vtkWidget.GetRenderWindow().AddRenderer(
                    self.vtk_renderer)
                self.visualizer = visu_cls(self.vtk_renderer)
                self.vtkWidget.Initialize()
            else:
                self._logger.warning("visualizer depends on vtk which is "
                                     "not available on this system!")
        else:
            raise NotImplementedError("Unsupported visualizer type '{}'"
                                      "".format(visu_cls))
        if hasattr(self, "actResetCamera"):
            self.actResetCamera.setEnabled(self.visualizer.can_reset_view)

    def add_plot_tree_item(self):
        text = "plot_{:03d}".format(self.dataPointTreeWidget.topLevelItemCount())
        if self._settings.value("plot/ask_for_plot_title", type=bool):
            name, ok = QInputDialog.getText(self,
                                            "Select a plot title",
                                            "name of the new plot:",
                                            text=text)
            if not (ok and name):
                return
        else:
            name = text

        similar_items = self.dataPointTreeWidget.findItems(name,
                                                           Qt.MatchExactly)
        if similar_items:
            self._logger.error("A plot with name '{}' already exists"
                               "".format(name))
            return

        if name in self.non_plotting_docks:
            self._logger.error("Name '{}' not allowed for a plot window "
                               "since it is already taken by another dock."
                               "".format(name))
            return

        top_level_item = QTreeWidgetItem()
        top_level_item.setText(0, name)
        self.dataPointTreeWidget.addTopLevelItem(top_level_item)
        top_level_item.setExpanded(1)
        self.dataPointTreeWidget.setCurrentItem(top_level_item)
        self.plot_vector_double_clicked(top_level_item)

    def show_selected_plot_tree_item(self):
        items = self.dataPointTreeWidget.selectedItems()
        for item in items:
            self.show_plot_tree_item(item)

    def remove_selected_plot_tree_items(self):
        items = self.dataPointTreeWidget.selectedItems()
        if not items:
            self._logger.error("Can't remove plot: no plot selected.")
            return

        for item in items:
            self.remove_plot_tree_item(item)

    def remove_plot_tree_item(self, item):
        # get the  top item
        while item.parent():
            item = item.parent()

        text = "The marked plot '" + item.text(0) + "' will be deleted!"
        reply_button = QMessageBox.warning(self, "Plot delete", text,
                                          QMessageBox.Ok | QMessageBox.Cancel)
        if reply_button == QMessageBox.Ok:
            open_docks = [dock.title() for dock in self.find_all_plot_docks()]
            if item.text(0) in open_docks:
                self.area.docks[item.text(0)].close()

            self.dataPointTreeWidget.takeTopLevelItem(
                self.dataPointTreeWidget.indexOfTopLevelItem(item))

    def add_data_point_to_tree(self):
        if not self.dataPointListWidget.selectedIndexes():
            self._logger.error("Can't add data set: no data set selected.")
            return

        data_points = []
        for item in self.dataPointListWidget.selectedItems():
            data_points.append(item.text())

        top_level_items = self.dataPointTreeWidget.selectedItems()
        if not top_level_items:
            if self.dataPointTreeWidget.topLevelItemCount() < 2:
                if self.dataPointTreeWidget.topLevelItemCount() < 1:
                    self.add_plot_tree_item()
                top_level_item = self.dataPointTreeWidget.topLevelItem(0)
            else:
                self._logger.error("Can't add data set: no plot selected.")
                return
        else:
            top_level_item = top_level_items[0]

        while top_level_item.parent():
            top_level_item = top_level_item.parent()

        top_level_item_list = []
        for i in range(top_level_item.childCount()):
            top_level_item_list.append(top_level_item.child(i).text(1))

        dock = next((d for d in self.find_all_plot_docks()
                     if d.title() == top_level_item.text(0)), None)

        for data_point in data_points:
            if data_point not in top_level_item_list:
                child = QTreeWidgetItem()
                child.setText(1, data_point)

                color = self._get_color(top_level_item.childCount())
                child.setBackground(0, color)

                top_level_item.addChild(child)

                if dock:
                    widget = dock.widgets[0]
                    self.plot_data_vector_member(child, widget)
            else:
                self._logger.error("Can't add data set: "
                                   "Set '{}' is already present selected plot"
                                   "".format(data_point))

    def export_data_points_from_tree(self):
        data_points = {}
        for item in self.dataPointListWidget.selectedItems():
            data_points[item.text()] = self._get_data_by_name(item.text())

        if not data_points:
            self._logger.error("Data export failed: No data selected.")
            return

        self.export(data_points)

    def remove_data_point_from_tree(self):
        items = self.dataPointTreeWidget.selectedItems()
        if not items:
            self._logger.error("Can't remove data set: no set selected.")
            return

        top_item = items[0]
        while top_item.parent():
            top_item = top_item.parent()

        top_item.takeChild(top_item.indexOfChild(items[0]))

        for i in range(top_item.childCount()):
            color_item = self._get_color(i)
            top_item.child(i).setBackground(0, color_item)

        self._update_plot(top_item)

    def _get_color(self, idx):
        sat_idx = idx % len(self.TABLEAU_COLORS)
        color = QColor(self.TABLEAU_COLORS[sat_idx][1])
        return color

    def _get_pen(self, idx):
        pen = pg.mkPen(self._get_color(idx), width=2)
        return pen

    def plot_vector_double_clicked(self, item):
        # check if a top level item has been clicked
        if not item.parent():
            self.show_plot_tree_item(item)

    def show_plot_tree_item(self, item):
        title = item.text(0)
        # check if plot has already been opened
        open_docks = [dock.title() for dock in self.find_all_plot_docks()]
        if title in open_docks:
            # update and raise the dock
            self._update_plot(item)
            try:
                self.area.docks[title].raiseDock()
            except:
                pass
        else:
            # create a new dock
            self.plot_data_vector(item)

    def load_last_sim(self, item):
        sim_name = str(item.text())
        try:
            idx = self.lastSimList.row(item)
        except ValueError as e:
            self._logger.error("load_last_sim(): No results called "
                               "'{0}'".format(sim_name))
            return False

        if idx >= len(self._lastSimulations):
            self._logger.error("load_last_sim(): Invalid index '{}')".format(idx))
            return False

        self._logger.info("restoring simulation '{}'".format(sim_name))

        self.currentDataset = self._lastSimulations[idx]
        if self._lastSimulations[idx]:
            self._read_results()
            self._update_data_list()
            self._update_plots()
            lsettings = self.currentDataset["modules"]
            self.visualizer.update_config(lsettings)
            lsettings["clear previous"] = True
            self.sim.restore_regime(lsettings)
            self.update_gui()

        self.setQListItemBold(self.lastSimList, item)
        self.setQListItemBold(self.regime_list, item)
        self.statusBar().showMessage(
            "restored simulation '{}'.".format(sim_name),
            1000)

    def _add_setting(self, setting, value):
        """
        Add a setting, if settings is present, no changes are made.

        Args:
            setting(str): Setting to add.
            value: Value to be set.
        """
        if not self._settings.contains(setting):
            self._settings.setValue(setting, value)

    def _init_settings(self):
        """
        Provide initial settings for the config file.

        """
        # path management
        self._add_setting("path/simulation_results",
                          os.path.join(os.path.curdir,
                                       "results",
                                       "simulation"))
        self._add_setting("path/simulation_results",
                          os.path.join(os.path.curdir,
                                       "results",
                                       "simulation"))
        self._add_setting("path/postprocessing_results",
                          os.path.join(os.path.curdir,
                                       "results",
                                       "postprocessing"))
        self._add_setting("path/metaprocessing_results",
                          os.path.join(os.path.curdir,
                                       "results",
                                       "metaprocessing"))
        self._add_setting("path/previous_plot_export", os.path.curdir)
        self._add_setting("path/previous_plot_format", ".csv")

        # control flow management
        self._add_setting("control/autoplay_animation", False)
        self._add_setting("control/exit_on_batch_completion", False)

        # view management
        self._add_setting("view/dock_state", {})
        self._add_setting("view/show_coordinates", True)
        self._add_setting("view/show_time_on_export", False)
        self._add_setting("view/export_width", "800")
        self._add_setting("view/export_height", "600")

        # log management
        self._add_setting("log_colors/CRITICAL", "#DC143C")
        self._add_setting("log_colors/ERROR", "#B22222")
        self._add_setting("log_colors/WARNING", "#DAA520")
        self._add_setting("log_colors/INFO", "#101010")
        self._add_setting("log_colors/DEBUG", "#4682B4")
        self._add_setting("log_colors/NOTSET", "#000000")

        # theme management
        self._add_setting("theme/use_dark_theme", False)

        # plot management
        self._add_setting("plot/ask_for_plot_title", False)

    def _write_settings(self):
        """ Store the application state. """
        pass

    @pyqtSlot()
    def update_autoplay_setting(self):
        self._settings.setValue("control/autoplay_animation",
                                str(self.actAutoPlay.isChecked()))

    @pyqtSlot()
    def update_show_coords_setting(self):
        self._settings.setValue("view/show_coordinates",
                                str(self.actShowCoords.isChecked()))

    @pyqtSlot()
    def update_exit_on_batch_completion_setting(self, state=None):
        if state is None:
            state = self.actExitOnBatchCompletion.isChecked()
        self._settings.setValue("control/exit_on_batch_completion", str(state))

    def set_visualizer(self, vis):
        self.visualizer = vis
        self.vtkWidget.Initialize()

    @pyqtSlot()
    def start_animation(self):
        """
        start the animation
        """
        self._logger.info("Starting Playback")

        # if we are at the end, start from the beginning
        if self.playbackTime >= self.currentEndTime:
            self.playbackTime = self.currentStartTime
            self.playbackTimeChanged.emit()

        self.actPlayPause.setText("Pause Animation")
        self.actPlayPause.setIcon(QIcon.fromTheme("pause"))
        self.actPlayPause.triggered.disconnect(self.start_animation)
        self.actPlayPause.triggered.connect(self.pause_animation)
        self.playbackTimer.start(self.playbackTimeout)

    @pyqtSlot()
    def pause_animation(self):
        """
        pause the animation
        """
        self._logger.info("Pausing Playback")
        self.playbackTimer.stop()
        self.actPlayPause.setText("Play Animation")
        self.actPlayPause.setIcon(QIcon.fromTheme("play"))
        self.actPlayPause.triggered.disconnect(self.pause_animation)
        self.actPlayPause.triggered.connect(self.start_animation)

    def stop_animation(self):
        """
        Stop the animation if it is running and reset the playback time.
        """
        self._logger.info("Stopping Playback")
        if self.actPlayPause.text() == "Pause Animation":
            # animation is playing -> stop it
            self.playbackTimer.stop()
            self.actPlayPause.setText("Play Animation")
            self.actPlayPause.setIcon(QIcon.fromTheme("play"))
            self.actPlayPause.triggered.disconnect(self.pause_animation)
            self.actPlayPause.triggered.connect(self.start_animation)

        self.playbackTime = self.currentStartTime
        self.playbackTimeChanged.emit()

    @pyqtSlot()
    def start_simulation(self):
        """
        start the simulation and disable start button
        """
        if self._current_regime_index is None:
            regime_name = ""
        else:
            regime_name = str(self.regime_list.item(
                self._current_regime_index).text())

        self.statusLabel.setText("simulating {}".format(regime_name))
        self._logger.info("Simulating: {}".format(regime_name))

        self.actSimulateCurrent.setIcon(QIcon(
            get_resource("stop_simulation")))
        self.actSimulateCurrent.setText("Abort &Simulation")
        self.actSimulateCurrent.triggered.disconnect(self.start_simulation)
        self.actSimulateCurrent.triggered.connect(self.stop_simulation)

        if not self.runningBatch:
            self.actSimulateAll.setDisabled(True)

        self.guiProgress = QProgressBar(self)
        self.sim.simulationProgressChanged.connect(self.guiProgress.setValue)
        self.statusBar().addWidget(self.guiProgress)
        self.runSimulation.emit()

    @pyqtSlot()
    def stop_simulation(self):
        self.stopSimulation.emit()

    def export_simulation_data(self, ok):
        """
        Query the user for a custom name and export the current simulation
        results.

        :param ok: unused parameter from QAction.triggered() Signal
        """
        self._save_data()

    def _save_data(self, file_path=None):
        """
        Save the current simulation results.

        If *file_path* is given, the result will be saved to the specified
        location, making automated exporting easier.

        Args:
            file_path(str): Absolute path of the target file. If `None` the
                use will be asked for a storage location.
        """
        regime_name = self._regimes[self._current_regime_index]["Name"]

        if file_path is None:
            # get default path
            path = self._settings.value("path/simulation_results")

            # create canonic file name
            suggestion = self._simfile_name(regime_name)
        else:
            path = os.path.dirname(file_path)
            suggestion = os.path.basename(file_path)

        # check if path exists otherwise create it
        if not os.path.isdir(path):
            box = QMessageBox()
            box.setText("Export Folder does not exist yet.")
            box.setInformativeText("Do you want to create it? \n"
                                   "{}".format(os.path.abspath(path)))
            box.setStandardButtons(QMessageBox.Ok | QMessageBox.No)
            box.setDefaultButton(QMessageBox.Ok)
            ret = box.exec_()
            if ret == QMessageBox.Ok:
                os.makedirs(path)
            else:
                path = os.path.abspath(os.path.curdir)
                file_path = None

        # If no path was given, present the default and let the user choose
        if file_path is None:
            dialog = QFileDialog(self)
            dialog.setAcceptMode(QFileDialog.AcceptSave)
            dialog.setFileMode(QFileDialog.AnyFile)
            dialog.setDirectory(path)
            dialog.setNameFilter("PyMoskito Results (*.pmr)")
            dialog.selectFile(suggestion)

            if dialog.exec_():
                file_path = dialog.selectedFiles()[0]
            else:
                self._logger.warning("Export Aborted")
                return -1

        # ask whether this should act as new default
        path = os.path.abspath(os.path.dirname(file_path))
        if path != self._settings.value("path/simulation_results"):
            box = QMessageBox()
            box.setText("Use this path as new default?")
            box.setInformativeText("{}".format(path))
            box.setStandardButtons(QMessageBox.Yes | QMessageBox.No)
            box.setDefaultButton(QMessageBox.Yes)
            ret = box.exec_()
            if ret == QMessageBox.Yes:
                self._settings.setValue("path/simulation_results", path)

        self.currentDataset.update({"regime name": regime_name})
        with open(file_path, "wb") as f:
            pickle.dump(self.currentDataset, f, protocol=4)

        self.statusLabel.setText("results saved to {}".format(file_path))
        self._logger.info("results saved to {}".format(file_path))

    def _simfile_name(self, regime_name):
        """ Create a canonical name for a simulation result file
        """
        suggestion = (time.strftime("%Y%m%d-%H%M%S")
                      + "_" + regime_name + ".pmr")
        return suggestion

    def load_regime_dialog(self):
        regime_path = os.path.join(os.curdir)

        dialog = QFileDialog(self)
        dialog.setFileMode(QFileDialog.ExistingFile)
        dialog.setDirectory(regime_path)
        dialog.setNameFilter("Simulation Regime files (*.sreg)")

        if dialog.exec_():
            file = dialog.selectedFiles()[0]
            self.load_regimes_from_file(file)

    def load_regimes_from_file(self, file_name):
        """
        load simulation regime from file
        :param file_name:
        """
        self.regime_file_name = os.path.split(file_name)[-1][:-5]
        self._logger.info("loading regime file: {0}".format(self.regime_file_name))
        with open(file_name.encode(), "r") as f:
            self._regimes += yaml.full_load(f)

        self._update_regime_list()

        if self._regimes:
            self.actSimulateAll.setDisabled(False)

        self._logger.info("loaded {} regimes".format(len(self._regimes)))
        self.statusBar().showMessage("loaded {} regimes.".format(len(self._regimes)), 1000)
        return

    def _update_regime_list(self):
        self.regime_list.clear()
        for reg in self._regimes:
            self._logger.info("adding '{}' to regime list".format(reg["Name"]))
            self.regime_list.addItem(reg["Name"])

    def remove_regime_items(self):
        if self.regime_list.currentRow() >= 0:
            # flag all selected files as invalid
            items = self.regime_list.selectedItems()
            for item in items:
                del self._regimes[self.regime_list.row(item)]
                self.regime_list.takeItem(self.regime_list.row(item))

    @pyqtSlot(QListWidgetItem)
    def regime_dclicked(self, item):
        """
        Apply the selected regime to the current target.
        """
        success = self._apply_regime_by_idx(self.regime_list.row(item))

        self.setQListItemBold(self.regime_list, item, success)
        self.setQListItemBold(self.lastSimList, item, success)
        self.dataPointListWidget.clear()

    def apply_regime_by_name(self, regime_name):
        """
        Apply the regime given by `regime_name` und update the regime index.

        Returns:
            bool: `True` if successful, `False` if errors occurred.
        """
        # get regime idx
        try:
            idx = list(map(itemgetter("Name"), self._regimes)).index(regime_name)
        except ValueError as e:
            self._logger.error("apply_regime_by_name(): Error no regime called "
                               "'{0}'".format(regime_name))
            return False

        # apply
        return self._apply_regime_by_idx(idx)

    def _apply_regime_by_idx(self, index=0):
        """
        Apply the given regime.

        Args:
            index(int): Index of the regime in the `RegimeList` .

        Returns:
            bool: `True` if successful, `False` if errors occurred.
        """
        if index >= len(self._regimes):
            self._logger.error("Invalid index: '{}'".format(index))
            return False

        reg_name = self._regimes[index]["Name"]
        self._logger.info("applying regime '{}'".format(reg_name))

        self._current_regime_index = index
        self._current_regime_name = reg_name

        ret = self.sim.set_regime(self._regimes[index])
        if ret:
            self.statusBar().showMessage("regime {} applied.".format(reg_name),
                                         1000)
        return ret

    @pyqtSlot()
    def start_regime_execution(self):
        """
        Simulate all regimes in the regime list.
        """
        self.actSimulateAll.setText("Stop Simulating &All Regimes")
        self.actSimulateAll.setIcon(QIcon.fromTheme("stop_batch"))
        self.actSimulateAll.triggered.disconnect(self.start_regime_execution)
        self.actSimulateAll.triggered.connect(self.stop_regime_execution)

        self.runningBatch = True
        self._current_regime_index = -1
        self.regimeFinished.emit()

    def run_next_regime(self):
        """
        Execute the next regime in the regime batch.
        """
        # are we finished?
        if self._current_regime_index == len(self._regimes) - 1:
            self.finishedRegimeBatch.emit(True)
            return

        suc = self._apply_regime_by_idx(self._current_regime_index + 1)
        if not suc:
            self.finishedRegimeBatch.emit(False)
            return

        self.start_simulation()

    @pyqtSlot()
    def stop_regime_execution(self):
        """ Stop the batch process.
        """
        self.stopSimulation.emit()
        self.finishedRegimeBatch.emit(False)

    def regime_batch_finished(self, status):
        self.runningBatch = False
        self.actSimulateAll.setDisabled(False)
        self.actSave.setDisabled(True)

        self.actSimulateAll.setText("Simulate &All Regimes")
        self.actSimulateAll.setIcon(QIcon.fromTheme("execute_regimes"))
        self.actSimulateAll.triggered.disconnect(self.stop_regime_execution)
        self.actSimulateAll.triggered.connect(self.start_regime_execution)

        if status:
            self.statusLabel.setText("All regimes have been simulated")
            self._logger.info("All Regimes have been simulated")
        else:
            self._logger.error("Batch simulation has been aborted")

        if self._settings.value("control/exit_on_batch_completion", type=bool):
            self._logger.info("Shutting down SimulationGUI")
            self.close()

    @pyqtSlot(str, dict, name="new_simulation_data")
    def new_simulation_data(self, status, data):
        """
        Slot to be called when the simulation interface has completed the
        current job and new data is available.

        Args:
            status (str): Status of the simulation, either
                - `finished` : Simulation has been finished successfully or
                - `failed` : Simulation has failed.
            data (dict): Dictionary, holding the simulation data.
        """
        self._logger.info("Simulation {}".format(status))
        self.statusLabel.setText("Simulation {}".format(status))

        self.actSimulateCurrent.setText("&Simulate Current Regime")
        self.actSimulateCurrent.setIcon(QIcon.fromTheme("simulate"))
        self.actSimulateCurrent.triggered.disconnect(self.stop_simulation)
        self.actSimulateCurrent.triggered.connect(self.start_simulation)

        self.actPlayPause.setDisabled(False)
        self.actStop.setDisabled(False)
        self.actSave.setDisabled(False)
        self.speedControl.setDisabled(False)
        self.timeSlider.setDisabled(False)

        self.sim.simulationProgressChanged.disconnect(self.guiProgress.setValue)
        self.statusBar().removeWidget(self.guiProgress)

        self.stop_animation()

        self.currentDataset = data

        # check if at least one time step of the simulation was computed
        if data and "results" in data and "Solver" in data["results"]:
            # import new data
            self._read_results()
            self._update_data_list()
            self._update_plots()
            self.visualizer.update_config(data["modules"])

            # add results to history
            lastSimCount = self.lastSimList.count()
            lastSimData = {'modules': data['modules'],
                           'results': data['results'],
                           'simulation': data['simulation'],
                           'name': self._current_regime_name,
                           }
            display_name = "{}:{}".format(lastSimCount,
                                          self._current_regime_name)
            self._lastSimulations.append(lastSimData)
            new_item = QListWidgetItem(display_name)
            self.lastSimList.addItem(new_item)
            self.lastSimList.scrollToItem(new_item)
            self.setQListItemBold(self.lastSimList, new_item)

        if self._settings.value("control/autoplay_animation", type=bool):
            self.actPlayPause.trigger()

        if self.runningBatch:
            regime_name = self._regimes[self._current_regime_index]["Name"]
            file_name = self._simfile_name(regime_name)
            self._save_data(os.path.join(
                self._settings.value("path/simulation_results"),
                file_name))
            self.regimeFinished.emit()
        else:
            self.actSimulateAll.setDisabled(False)

    def _read_results(self):
        sources = ["Limiter", "ModelMixer", "Controller", "Feedforward"]
        input_source = next(s for s in sources if s in self.currentDataset["results"])
        model_input = self.currentDataset["results"][input_source]
        self.inputInterpolator = interp1d(self.currentDataset["results"]["time"],
                                          model_input,
                                          axis=0,
                                          bounds_error=False,
                                          fill_value=(model_input[0], model_input[-1]))
        state = self.currentDataset["results"]["Solver"]
        self.stateInterpolator = interp1d(self.currentDataset["results"]["time"],
                                          state,
                                          axis=0,
                                          bounds_error=False,
                                          fill_value=(state[0], state[-1]))
        self.currentStepSize = 1.0 / self.currentDataset["simulation"][
            "measure rate"]
        self.currentStartTime = self.currentDataset["simulation"]["start time"]
        self.currentEndTime = self.currentDataset["simulation"]["end time"]

    def increment_playback_speed(self):
        self.speedControl.setValue(self.speedControl.value()
                                   + self.speedControl.singleStep())

    def decrement_playback_speed(self):
        self.speedControl.setValue(self.speedControl.value()
                                   - self.speedControl.singleStep())

    def reset_playback_speed(self):
        self.speedControl.setValue((self.speedControl.maximum()
                                    - self.speedControl.minimum()) / 2)

    def set_slowest_playback_speed(self):
        self.speedControl.setValue(self.speedControl.minimum())

    def set_fastest_playback_speed(self):
        self.speedControl.setValue(self.speedControl.maximum())

    def update_playback_speed(self, val):
        maximum = self.speedControl.maximum()
        self.playbackGain = 10 ** (3.0 * (val - maximum / 2) / maximum)

    @pyqtSlot()
    def increment_playback_time(self):
        """
        go one time step forward in playback
        """
        if self.playbackTime >= self.currentEndTime:
            self.pause_animation()
            return

        increment = self.playbackGain * self.playbackTimeout / 1000
        self.playbackTime = min(self.currentEndTime,
                                self.playbackTime + increment)
        self.playbackTimeChanged.emit()

    def _update_time_slider(self):
        time_delta = self.currentEndTime - self.currentStartTime
        if time_delta == 0:
            # no meaningful context to update the slider
            return

        rel_pos = (self.playbackTime - self.currentStartTime) / (
            time_delta)
        pos = int(rel_pos * self.timeSliderRange)
        self.timeSlider.blockSignals(True)
        self.timeSlider.setValue(pos)
        self.timeSlider.blockSignals(False)

    def update_playback_time(self):
        """
        adjust playback time to slider value
        """
        self.playbackTime = self.timeSlider.value() / self.timeSliderRange * (
                self.currentEndTime
                - self.currentStartTime
        ) + self.currentStartTime
        self.playbackTimeChanged.emit()

    def update_gui(self):
        """
        Updates the graphical user interface to reflect changes of the current
        display time. 
        
        This includes:
            - timestamp
            - visualisation window
            - time cursors in diagrams
        """
        self.timeLabel.setText("t={0:.3e}".format(self.playbackTime))

        # update timing elements
        self._update_time_slider()
        self._update_time_cursors()

        # update state of rendering
        if self.visualizer is not None and self.stateInterpolator is not None:
            inp = self.inputInterpolator(self.playbackTime)
            state = self.stateInterpolator(self.playbackTime)
            self.visualizer.update_scene(state, inp)
            if isinstance(self.visualizer, MplVisualizer):
                pass
            elif isinstance(self.visualizer, VtkVisualizer):
                self.vtkWidget.GetRenderWindow().Render()

    def _update_data_list(self):
        # self.dataList.clear()
        self.dataPointListWidget.clear()
        # TODO lets open and check if possible to plot
        # TODO create trees with children instead of plain suffixes
        for module_name, results in self.currentDataset["results"].items():
            if not isinstance(results, np.ndarray):
                continue
            if len(results.shape) == 1:
                self.dataPointListWidget.addItem(module_name)
            elif len(results.shape) == 2:
                for col in range(results.shape[1]):
                    self.dataPointListWidget.addItem(
                        self._build_entry_name(module_name, (col,))
                    )
            elif len(results.shape) == 3:
                for col in range(results.shape[1]):
                    for der in range(results.shape[2]):
                        self.dataPointListWidget.addItem(
                            self._build_entry_name(module_name, (col, der))
                        )

    def _build_entry_name(self, module_name, idx):
        """
        Construct an identifier for a given entry of a module.
        Args:
            module_name (str): name of the module the entry belongs to.
            idx (tuple): Index of the entry.

        Returns:
            str: Identifier to use for display.
        """
        # save the user from defining 1d entries via tuples
        if len(idx) == 1:
            m_idx = idx[0]
        else:
            m_idx = idx

        mod_settings = self.currentDataset["modules"]
        info = mod_settings.get(module_name, {}).get("output_info", None)
        if info:
            if m_idx in info:
                return ".".join([module_name, info[m_idx]["Name"]])

        return ".".join([module_name] + [str(i) for i in idx])

    def _get_index_from_suffix(self, module_name, suffix):
        info = self.currentDataset["modules"].get(module_name, {}).get(
            "output_info", None)
        idx = next((i for i in info if info[i]["Name"] == suffix), None)
        return idx

    def _get_units(self, entry):
        """
        Return the unit that corresponds to a given entry.

        If no information is available, None is returned.

        Args:
            entry (str): Name of the entry. This can either be "Model.a.b" where
                a and b are numbers or if information is available "Model.Signal"
                where signal is the name of that part.

        Returns:

        """
        args = entry.split(".")
        module_name = args.pop(0)
        info = self.currentDataset["modules"].get(module_name, {}).get(
            "output_info", None)
        if info is None:
            return None

        if len(args) == 1:
            try:
                idx = int(args[0])
            except ValueError:
                idx = next((i for i in info if info[i]["Name"] == args[0]),
                           None)
        else:
            idx = (int(a) for a in args)

        return info[idx]["Unit"]

    def _update_plot(self, item):
        # collect data
        if self.currentDataset is None:
            return

        title = item.text(0)
        t = self.currentDataset["results"]["time"]
        docks = self.find_all_plot_docks()
        target = next((d for d in docks if d.title() == title), None)
        if target is None:
            return

        for widget in target.widgets:
            child_names = [item.child(c_idx).text(1)
                           for c_idx in range(item.childCount())]
            del_list = []
            cnt = 0
            for _item in widget.getPlotItem().items:
                if isinstance(_item, pg.PlotDataItem):
                    if _item.name() in child_names:
                        y_data = self._get_data_by_name(_item.name())
                        if y_data is not None:
                            _item.setData(x=t, y=y_data)
                            _item.setPen(self._get_pen(cnt))
                            cnt += 1
                        else:
                            _item.clear()
                    else:
                        del_list.append(_item)

            for _item in del_list:
                widget.getPlotItem().removeItem(_item)

    def plot_data_vector(self, item):
        """
        Creates a plot widget based on the given item.

        If a plot for this item is already open no new plot is created but the
        existing one is raised up again.

        Args:
            item(Qt.ListItem): Item to plot.
        """
        if self.currentDataset is None:
            return

        title = str(item.text(0))

        # create plot widget
        widget = pg.PlotWidget()
        widget.showGrid(True, True)
        widget.getPlotItem().getAxis("bottom").setLabel(text="Time", units="s")

        for idx in range(item.childCount()):
            self.plot_data_vector_member(item.child(idx), widget)

        # add a time line
        time_line = pg.InfiniteLine(self.playbackTime,
                                    angle=90,
                                    movable=False,
                                    pen=pg.mkPen("#FF0000", width=2.0))
        widget.getPlotItem().addItem(time_line)

        coord_item = pg.TextItem(text='', anchor=(0, 1))
        widget.getPlotItem().addItem(coord_item, ignoreBounds=True)

        def _info_wrapper(pos):
            self.update_coord_info(pos, widget, coord_item)

        widget.scene().sigMouseMoved.connect(_info_wrapper)

        # add custom export entries
        widget.scene().contextMenu = [
            QAction("Export as ...", self),
        ]

        def _export_wrapper(export_func):
            def _wrapper():
                return export_func(widget.getPlotItem(),
                                   title,
                                   coord_item,
                                   time_line)
            return _wrapper

        widget.scene().contextMenu[0].triggered.connect(
            _export_wrapper(self.export_plot_item))

        # create dock container and add it to dock area
        dock = pg.dockarea.Dock(title, closable=True)
        dock.addWidget(widget)

        plotWidgets = self.find_all_plot_docks()
        if plotWidgets:
            self.area.addDock(dock, "above", plotWidgets[0])
        else:
            self.area.addDock(dock, "bottom", self.animationDock)

    def update_coord_info(self, pos, widget, coord_item):
        mouse_coords = widget.getPlotItem().vb.mapSceneToView(pos)
        coord_item.setPos(mouse_coords.x(), mouse_coords.y())
        coord_text = "x={:.3e} y={:.3e}".format(mouse_coords.x(),
                                                mouse_coords.y())
        self.coordLabel.setText(coord_text)

        show_info = self._settings.value("view/show_coordinates", type=bool)
        if widget.sceneBoundingRect().contains(pos) and show_info:
            coord_item.setText(coord_text.replace(" ", "\n"))
            coord_item.show()
        else:
            coord_item.hide()

    def plot_data_vector_member(self, item, widget):
        idx = item.parent().indexOfChild(item)

        data_name = item.text(1)
        t = self.currentDataset["results"]["time"]
        data = self._get_data_by_name(data_name)
        if data is None:
            t = None

        widget.plot(x=t,
                    y=data,
                    pen=self._get_pen(idx),
                    name=data_name)

    def find_all_plot_docks(self):
        list = []
        for title, dock in self.area.findAll()[1].items():
            if title in self.non_plotting_docks:
                continue
            else:
                list.append(dock)

        return list

    def export_plot_item(self, plot_item, name, coord_item, time_item):
        plot_data = {}
        for i, c in enumerate(plot_item.curves):
            if c.getData() is None:
                continue
            if len(c.getData()) > 2:
                self._logger.warning("Can not handle data format of entry"
                                     "'{}'!".format(c.name()))
                continue
            if "time" not in plot_data:
                plot_data["time"] = c.getData()[0]

            c_name = c.name()
            c_unit = self._get_units(c_name)
            if c_unit is not None:
                c_name += " ({})".format(c_unit)

            plot_data[c_name] = c.getData()[1]

        self.export(plot_data)

    def export(self, plot_data):
        try:
            exporter = Exporter(data_points=plot_data)
        except Exception as e:
            self._logger.error("Can't instantiate exporter! " + str(e))
            return

        last_path = self._settings.value("path/previous_plot_export")
        last_format = self._settings.value("path/previous_plot_format")
        export_formats = ["CSV Data (*.csv)", "PNG Image (*.png)"]
        if last_format == ".png":
            export_formats[:] = export_formats[::-1]
        format_str = ";;".join(export_formats)
        default_file = os.path.join(last_path, "export" + last_format)
        filename = QFileDialog.getSaveFileName(self,
                                               "Export as ...",
                                               default_file,
                                               format_str)

        if filename[0]:
            file, ext = os.path.splitext(filename[0])
            self._settings.setValue("path/previous_plot_export",
                                    os.path.dirname(file))
            if ext == '.csv':
                exporter.export_csv(filename[0])
                self._settings.setValue("path/previous_plot_format", ".csv")
            elif ext == '.png':
                exporter.export_png(filename[0])
                self._settings.setValue("path/previous_plot_format", "")
            else:
                self._logger.error("Wrong extension used!")
                return
            self._logger.info("Export successful as '{}'.".format(filename[0]))

    def _get_data_by_name(self, name):
        tmp = name.split(".")
        module_name = tmp[0]
        try:
            raw_data = self.currentDataset["results"][module_name]
        except KeyError:
            return None

        if len(tmp) == 1:
            data = np.array(raw_data)
        elif len(tmp) == 2:
            try:
                idx = int(tmp[1])
            except ValueError:
                idx = self._get_index_from_suffix(module_name, tmp[1])
            if raw_data.ndim != 2 or raw_data.shape[1] <= idx:
                return None
            data = raw_data[:, idx]
        elif len(tmp) == 3:
            try:
                idx = int(tmp[1])
                der = int(tmp[2])
            except ValueError:
                return None
            if raw_data.ndim != 3 or raw_data.shape[1] <= idx \
                    or raw_data.shape[2] <= der:
                return None
            data = raw_data[:, idx, der]
        else:
            raise ValueError("Format not supported")

        return data

    def _update_time_cursors(self):
        """
        Update the time lines of all plot windows
        """
        for title, dock in self.area.findAll()[1].items():
            if title in self.non_plotting_docks:
                continue
            for widget in dock.widgets:
                for item in widget.getPlotItem().items:
                    if isinstance(item, pg.InfiniteLine):
                        item.setValue(self.playbackTime)

    def _update_plots(self):
        """
        Update the data in all plot windows
        """
        root = self.dataPointTreeWidget.invisibleRootItem()
        for i in range(root.childCount()):
            self._update_plot(root.child(i))

    @pyqtSlot(QModelIndex)
    def target_view_changed(self, index):
        self.targetView.resizeColumnToContents(0)

    def postprocessing_clicked(self):
        """
        starts the post- and metaprocessing application
        """
        self._logger.info("launching postprocessor")
        self.statusBar().showMessage("launching postprocessor", 1000)
        if self.postprocessor is None:
            self.postprocessor = PostProcessor(self._settings)

        self.postprocessor.show()

    def reset_camera_clicked(self):
        """
        reset camera in vtk window
        """
        self.visualizer.reset_camera()
        self.vtkWidget.GetRenderWindow().Render()

    def show_info(self):
        icon_lic = open(get_resource("license.txt"), "r").read()
        text = "This application was build using PyMoskito ver. {} .<br />" \
               "PyMoskito is free software distributed under GPLv3. <br />" \
               "It is currently developed by members of the " \
               "<a href=\'https://www.umit-tirol.at/page.cfm?vpath=departments/technik/iace&switchLocale=en_US'>" \
               "Institute of Automation and Control Engineering</a>" \
               " at <a href=\'https://www.umit-tirol.at'>" \
               "UMIT-Triol</a>. <br />" \
               "".format(metadata.version("pymoskito")) \
               + "<br />" + icon_lic
        box = QMessageBox.about(self, "PyMoskito", text)

    def show_online_docs(self):
        webbrowser.open("https://pymoskito.readthedocs.org")

    def closeEvent(self, QCloseEvent):
        self._logger.info("Close Event received, shutting down.")
        logging.getLogger().removeHandler(self.textLogger)
        self._settings.setValue("view/dock_state", self.area.saveState())
        super().closeEvent(QCloseEvent)

    def loadStandardDockState(self):
        for docks in self.find_all_plot_docks():
            docks.close()
        self.area.restoreState(self.standardDockState)

    def setQListItemBold(self, q_list=None, item=None, state=True):
        for i in range(q_list.count()):
            new_font = q_list.item(i).font()
            if q_list.item(i) == item and state:
                new_font.setBold(1)
            else:
                new_font.setBold(0)
            q_list.item(i).setFont(new_font)
        q_list.repaint()

    def toggleTheme(self):
        darkmode = not self._settings.value("theme/use_dark_theme", type=bool)
        self.setTheme(darkmode)
        self._settings.setValue("theme/use_dark_theme", darkmode)

    def setTheme(self, use_dark_theme=None):
        if use_dark_theme is None:
            use_dark_theme = self._settings.value("theme/use_dark_theme", type=bool)

        if use_dark_theme:
            QIcon.setThemeName("dark")
            QApplication.instance().setStyle(DarkModeStyle())
            QApplication.instance().setPalette(darkPalette)
            self._settings.setValue("log_colors/INFO", "#efefef")
            self._settings.setValue("log_colors/NOTSET", "#fffff")
        else:
            QIcon.setThemeName("light")
            QApplication.instance().setStyle(LightModeStyle())
            QApplication.instance().setPalette(lightPalette)
            self._settings.setValue("log_colors/INFO", "#101010")
            self._settings.setValue("log_colors/NOTSET", "#00000")

        if hasattr(self, "area"):
            for dock in self.area.docks.values():
                dock.updateStyle()

        if hasattr(self, "visualizer"):
            self.visualizer.update_theme()



from PyQt5.QtWidgets import QStyle, QProxyStyle
from PyQt5.QtGui import QPalette, QColor, QIcon, QPainter

# as light is the default, just remember that
lightPalette =  QPalette()

darkPalette = QPalette()

bg_color = QColor(53, 53, 53)
fg_color = QColor(239, 239, 239)
disabled_color = QColor(127, 127, 127)
ac_color1 = QColor("#00FF85")  # neon green
ac_color2 = QColor("#1E90FF")  # electric blue
ac_color3 = QColor("#FF0099")  # vivid pink

darkPalette.setColor(QPalette.ColorRole.Window, bg_color)
darkPalette.setColor(QPalette.ColorRole.WindowText, fg_color)
darkPalette.setColor(
    QPalette.ColorGroup.Disabled,
    QPalette.ColorRole.WindowText,
    disabled_color,
)
darkPalette.setColor(QPalette.ColorRole.Base, QColor(25, 25, 25))
darkPalette.setColor(QPalette.ColorRole.AlternateBase, bg_color)

darkPalette.setColor(QPalette.ColorRole.ToolTipBase, bg_color)
darkPalette.setColor(QPalette.ColorRole.ToolTipText, fg_color)

# text
darkPalette.setColor(QPalette.ColorRole.Text, fg_color)
darkPalette.setColor(
    QPalette.ColorGroup.Disabled,
    QPalette.ColorRole.Text,
    disabled_color,
)
darkPalette.setColor(QPalette.ColorRole.BrightText, ac_color1)
darkPalette.setColor(
    QPalette.ColorGroup.Disabled,
    QPalette.ColorRole.Highlight,
    disabled_color,
)
darkPalette.setColor(QPalette.ColorRole.HighlightedText, fg_color)
darkPalette.setColor(QPalette.ColorRole.PlaceholderText, fg_color)

# buttons
darkPalette.setColor(QPalette.ColorRole.Button, bg_color)
darkPalette.setColor(QPalette.ColorRole.ButtonText, fg_color)
darkPalette.setColor(
    QPalette.ColorGroup.Disabled,
    QPalette.ColorRole.ButtonText,
    disabled_color,
)

# special roles
darkPalette.setColor(QPalette.ColorRole.Link, ac_color1)
darkPalette.setColor(QPalette.ColorRole.Highlight, ac_color2)
darkPalette.setColor(QPalette.ColorRole.Dark, QColor(35, 35, 35, 35))
darkPalette.setColor(QPalette.ColorRole.Shadow, QColor(20, 20, 20, 20))


class LightModeStyle(QProxyStyle):
    pass


class DarkModeStyle(QProxyStyle):

    def generatedIconPixmap(self, iconMode, pixmap, opt):
        """ Drop in replacement for modal icon rendering """
        if iconMode == QIcon.Mode.Disabled:
            painter = QPainter(pixmap)
            painter.setCompositionMode(QPainter.CompositionMode_SourceIn)
            painter.fillRect(pixmap.rect(), disabled_color)
            painter.end()
            return pixmap
        return super().generatedIconPixmap(iconMode, pixmap, opt)
